# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

import logging
import socket

from bagpipe.bgp.common import utils
from bagpipe.bgp.common import logDecorator

from bagpipe.bgp.engine import RouteEvent

from bagpipe.bgp.vpn.vpn_instance import VPNInstance
from bagpipe.bgp.vpn.dataplane_drivers import \
    DummyDataplaneDriver as _DummyDataplaneDriver
from bagpipe.bgp.vpn.dataplane_drivers import \
    VPNInstanceDataplane as _VPNInstanceDataplane
from bagpipe.bgp.vpn.dataplane_drivers import \
    DummyVPNInstanceDataplane as _DummyVPNInstanceDataplane

from bagpipe.bgp.common.looking_glass import LookingGlass, LGMap, \
    LookingGlassReferences

from bagpipe.exabgp.structure.vpn import RouteDistinguisher
from bagpipe.exabgp.structure.evpn import EVPNNLRI, EVPNMACAdvertisement, \
    EVPNMulticast, EthernetSegmentIdentifier, EthernetTag, MAC
from bagpipe.exabgp.structure.mpls import LabelStackEntry
from bagpipe.exabgp.structure.address import AFI, SAFI
from bagpipe.exabgp.structure.ip import Inet
from bagpipe.exabgp.message.update.route import Route
from bagpipe.exabgp.message.update.attribute.nexthop import NextHop
from bagpipe.exabgp.message.update.attribute.communities import Encapsulation
from bagpipe.exabgp.message.update.attribute.pmsi_tunnel import PMSITunnel, \
    PMSITunnelIngressReplication
from bagpipe.exabgp.message.update.attribute.id import AttributeID


class VPNInstanceDataplane(_VPNInstanceDataplane):
    __metaclass__ = ABCMeta

    @abstractmethod
    def addDataplaneForBroadcastEndpoint(self, remotePE, label, nlri, encaps):
        pass

    @abstractmethod
    def removeDataplaneForBroadcastEndpoint(self, remotePE, label, nlri):
        pass

    @abstractmethod
    def setGatewayPort(self, linuxif):
        '''
        Used to determine a port to which traffic at the destination of the
        IP gateway should be sent.  This is used to plug an EVI into an IP VPN
        VRF.
        '''
        pass

    @abstractmethod
    def gatewayPortDown(self, linuxif):
        '''
        Used to revert the action done when setGatewayPort was called.
        Relevant only when an EVI had been plugged into an IP VPN VRF.
        '''
        pass


class DummyVPNInstanceDataplane(_DummyVPNInstanceDataplane,
                                _VPNInstanceDataplane):
    '''
    Dummy, do-nothing dataplane driver
    '''

    @logDecorator.logInfo
    def addDataplaneForBroadcastEndpoint(self, remotePE, label, nlri, encaps):
        pass

    @logDecorator.logInfo
    def removeDataplaneForBroadcastEndpoint(self, remotePE, label, nlri):
        pass

    @logDecorator.logInfo
    def setGatewayPort(self, linuxif):
        pass

    @logDecorator.logInfo
    def gatewayPortDown(self, linuxif):
        pass


class DummyDataplaneDriver(_DummyDataplaneDriver):

    dataplaneInstanceClass = DummyVPNInstanceDataplane
    encaps = [Encapsulation(Encapsulation.VXLAN)]

    def __init__(self, *args):
        _DummyDataplaneDriver.__init__(self, *args)


# EVI

class EVI(VPNInstance, LookingGlass):

    '''
    Implementation an E-VPN instance (EVI) based on RFC7432 and
    draft-ietf-bess-evpn-overlay.
    '''

    type = "evpn"
    afi = AFI(AFI.l2vpn)
    safi = SAFI(SAFI.evpn)

    @logDecorator.log
    def __init__(self, *args, **kwargs):

        VPNInstance.__init__(self, *args, **kwargs)

        self.gwPort = None

        # Advertise route to receive multi-destination traffic
        self.log.info("Generating BGP route for broadcast/multicast traffic")

        etag = None
        label = LabelStackEntry(self.instanceLabel)

        route = Route(
            EVPNMulticast(
                RouteDistinguisher(RouteDistinguisher.TYPE_IP_LOC,
                                   None,
                                   self.bgpManager.getLocalAddress(),
                                   self.instanceId),
                etag,
                self.bgpManager.getLocalAddress()
            )
        )

        route.attributes.add(self._genExtendedCommunities())

        # add PMSI Tunnel Attribute route
        pmsi_tunnel_attribute = PMSITunnelIngressReplication(
            self.dataplaneDriver.getLocalAddress(), label)
        route.attributes.add(pmsi_tunnel_attribute)

        nh = Inet(1, socket.inet_pton(
            socket.AF_INET, self.dataplaneDriver.getLocalAddress()))
        route.attributes.add(NextHop(nh))

        self.multicastRouteEntry = self._newRouteEntry(self.afi,
                                                       self.safi,
                                                       self.exportRTs,
                                                       route.nlri,
                                                       route.attributes)

        self._pushEvent(
            RouteEvent(RouteEvent.ADVERTISE, self.multicastRouteEntry))

    def generateVifBGPRoute(self, macAddress, ipPrefix, prefixLen, label):
        # Generate BGP route and advertise it...

        assert(prefixLen == 32)

        # label parameter ignored, we want the instance label
        lse = LabelStackEntry(self.instanceLabel, True)
        etag = None

        route = Route(
            EVPNMACAdvertisement(
                RouteDistinguisher(RouteDistinguisher.TYPE_IP_LOC, None,
                                   self.dataplaneDriver.getLocalAddress(),
                                   self.instanceId),
                EthernetSegmentIdentifier(),
                etag,
                MAC(macAddress),
                lse,
                ipPrefix
            )
        )

        return self._newRouteEntry(self.afi, self.safi, self.exportRTs,
                                   route.nlri, route.attributes)

    @logDecorator.log
    def setGatewayPort(self, linuxif, ipvpn):
        self.dataplane.setGatewayPort(linuxif)
        self.gwPort = (linuxif, ipvpn)

    @logDecorator.log
    def gatewayPortDown(self, linuxif):
        self.dataplane.gatewayPortDown(linuxif)
        self.gwPort = None

    def hasGatewayPort(self):
        return (self.gwPort is not None)

    # TrackerWorker callbacks for BGP route updates ##########################

    def _route2trackedEntry(self, route):
        if isinstance(route.nlri, EVPNMACAdvertisement):
            return (EVPNMACAdvertisement, route.nlri.mac)
        elif isinstance(route.nlri, EVPNMulticast):
            return (EVPNMulticast, (route.nlri.ip, route.nlri.rd))
        elif isinstance(route.nlri, EVPNNLRI):
            self.log.warning("Received EVPN route of unsupported subtype: %s",
                             route.nlri.subtype)
        else:
            raise Exception("EVI %d should not receive routes of type %s" %
                            (self.instanceId, type(route.nlri)))

    @utils.synchronized
    @logDecorator.log
    def _newBestRoute(self, entry, newRoute):
        (entryClass, info) = entry

        encaps = self._checkEncaps(newRoute)
        if not encaps:
            return

        if entryClass == EVPNMACAdvertisement:
            prefix = info

            nh = newRoute.attributes.get(NextHop.ID)
            remotePE = nh.next_hop
            label = newRoute.nlri.label.labelValue

            self.dataplane.setupDataplaneForRemoteEndpoint(
                prefix, remotePE, label, newRoute.nlri, encaps)

        elif entryClass == EVPNMulticast:
            remote_endpoint = info

            # check that the route is actually carrying an PMSITunnel of type
            # ingress replication
            pmsi_tunnel = newRoute.attributes.get(PMSITunnel.ID)
            if not isinstance(pmsi_tunnel, PMSITunnelIngressReplication):
                self.log.warning("Received PMSITunnel of unsupported type: %s",
                                 type(pmsi_tunnel))
            else:
                remote_endpoint = pmsi_tunnel.ip
                label = pmsi_tunnel.label.labelValue

                self.log.info("Setting up dataplane for new ingress "
                              "replication destination %s", remote_endpoint)
                self.dataplane.addDataplaneForBroadcastEndpoint(
                    remote_endpoint, label, newRoute.nlri, encaps)
        else:
            self.log.warning("unsupported entryClass: %s", entryClass.__name__)

    @utils.synchronized
    @logDecorator.log
    def _bestRouteRemoved(self, entry, oldRoute, last):
        (entryClass, info) = entry

        if entryClass == EVPNMACAdvertisement:

            if self._skipRouteRemoval(last):
                self.log.debug("Skipping removal of non-last route because "
                               "dataplane does not want it")
                return

            prefix = info

            nh = oldRoute.attributes.get(NextHop.ID)
            remotePE = nh.next_hop
            label = oldRoute.nlri.label.labelValue

            self.dataplane.removeDataplaneForRemoteEndpoint(
                prefix, remotePE, label, oldRoute.nlri)

        elif entryClass == EVPNMulticast:
            remote_endpoint = info

            # check that the route is actually carrying an PMSITunnel of type
            # ingress replication
            pmsi_tunnel = oldRoute.attributes.get(PMSITunnel.ID)
            if not isinstance(pmsi_tunnel, PMSITunnelIngressReplication):
                self.log.warning("PMSITunnel of suppressed route is of"
                                 " unsupported type")
            else:
                remote_endpoint = pmsi_tunnel.ip
                label = pmsi_tunnel.label.labelValue
                self.log.info("Cleaning up dataplane for ingress replication "
                              "destination %s", remote_endpoint)
                self.dataplane.removeDataplaneForBroadcastEndpoint(
                    remote_endpoint, label, oldRoute.nlri)
        else:
            self.log.warning("unsupported entryClass: %s", entryClass.__name__)

    # Looking Glass ####

    def getLookingGlassLocalInfo(self, pathPrefix):
        if not self.gwPort:
            return {"gateway_port": None}
        else:
            (linuxif, ipvpn) = self.gwPort
            return {"gateway_port": {
                    "interface": repr(linuxif),
                    "ipvpn": {"href":
                              LookingGlassReferences.getAbsolutePath(
                                  "VPN_INSTANCES", pathPrefix,
                                  [ipvpn.externalInstanceId]),
                              "id": ipvpn.name,
                              "external_instance_id":
                                  ipvpn.externalInstanceId
                              },
                    }}



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import socket

import logging

from bagpipe.bgp.engine.route_table_manager import RouteTableManager, \
    WorkerCleanupEvent
from bagpipe.bgp.engine.bgp_peer_worker import BGPPeerWorker
from bagpipe.bgp.engine.exabgp_peer_worker import ExaBGPPeerWorker
from bagpipe.bgp.engine import RouteEvent, RouteEntry, \
    Subscription, Unsubscription

from bagpipe.bgp.common.looking_glass import LookingGlass, LGMap
from bagpipe.bgp.common.utils import getBoolean
from bagpipe.bgp.common import logDecorator

from bagpipe.exabgp.message.update.route import Route
from bagpipe.exabgp.structure.rtc import RouteTargetConstraint
from bagpipe.exabgp.structure.address import AFI, SAFI
from bagpipe.exabgp.structure.ip import Inet
from bagpipe.exabgp.message.update.attribute.nexthop import NextHop

log = logging.getLogger(__name__)


class Manager(LookingGlass):

    def __init__(self, _config, peerClass=ExaBGPPeerWorker):
        log.debug("Instantiating Manager")

        self.config = _config
        self.peerClass = peerClass

        # RTC is defaults to being enabled
        self.config['enable_rtc'] = getBoolean(self.config.get('enable_rtc',
                                                               True))

        self.routeTableManager = RouteTableManager()
        self.routeTableManager.start()

        if 'local_address' not in self.config:
            raise Exception("config needs a local_address")

        if 'my_as' not in self.config:
            raise Exception("config needs a my_as")
        self.config['my_as'] = int(self.config['my_as'])

        if 'peer_as' in self.config:
            raise Exception("config must omit peer_as, because only iBGP "
                            "is supported yet")
        self.config['peer_as'] = self.config['my_as']

        self.peers = {}
        if self.config['peers']:
            peersAddresses = [x.strip() for x in
                              self.config['peers'].strip().split(",")]
            for peerAddress in peersAddresses:
                log.debug("Creating a peer worker for %s", peerAddress)
                peerWorker = self.peerClass(
                    self, None, peerAddress, self.config)
                self.peers[peerAddress] = peerWorker
                peerWorker.start()

        self.trackedSubs = dict()

        # we need a .name since we'll masquerade as a routeEntry source
        self.name = "BGPManager"

    @logDecorator.log
    def stop(self):
        for peer in self.peers.itervalues():
            peer.stop()
        self.routeTableManager.stop()
        for peer in self.peers.itervalues():
            peer.join()
        self.routeTableManager.join()

    def _pushEvent(self, routeEvent):
        log.debug("push event to RouteTableManager")
        self.routeTableManager.enqueue(routeEvent)

    def cleanup(self, worker):
        log.debug("push cleanup event for worker %s to RouteTableManager",
                  worker.name)
        self.routeTableManager.enqueue(WorkerCleanupEvent(worker))

        # TODO(tmmorin): withdraw RTC routes corresponding to worker
        # subscriptions -- currently ok since VPNInstance._stop() calls
        # unsubscribe

    def getLocalAddress(self):
        try:
            return self.config['local_address']
        except KeyError:
            log.error("BGPManager config has no localAddress defined")
            return "0.0.0.0"

    def routeEventSubUnsub(self, subobj):
        if isinstance(subobj, Subscription):
            self._routeEventSubscribe(subobj)
        elif isinstance(subobj, Unsubscription):
            self._routeEventUnsubscribe(subobj)
        else:
            assert(False)

    @logDecorator.log
    def _routeEventSubscribe(self, subscription):

        self.routeTableManager.enqueue(subscription)

        # synthesize a RouteEvent for a RouteTarget constraint route
        if (self.config['enable_rtc'] and not isinstance(subscription.worker,
                                                         BGPPeerWorker)):

            firstWorkerForSubscription = self._trackedSubscriptionsAddWorker(
                subscription)

            # FIXME: not excellent to hardcode this here
            if ((subscription.safi in (SAFI.mpls_vpn, SAFI.evpn))
                    and firstWorkerForSubscription):
                routeEvent = RouteEvent(
                    RouteEvent.ADVERTISE,
                    self._subscription2RTCRouteEntry(subscription), self)
                log.debug(
                    "Based on subscription => synthesized RTC %s", routeEvent)
                self.routeTableManager.enqueue(routeEvent)
            else:
                log.debug("No need to synthesize an RTC route "
                          "(firstWorkerForSubscription:%s) ",
                          firstWorkerForSubscription)

    @logDecorator.log
    def _routeEventUnsubscribe(self, unsubscription):

        self.routeTableManager.enqueue(unsubscription)

        if (self.config['enable_rtc'] and not isinstance(unsubscription.worker,
                                                         BGPPeerWorker)):

            wasLastWorkerForSubscription = \
                self._trackedSubscriptionsRemoveWorker(unsubscription)

            # FIXME: not excellent to hardcode this here
            if ((unsubscription.safi in (SAFI.mpls_vpn, SAFI.evpn))
                    and wasLastWorkerForSubscription):
                # synthesize a withdraw RouteEvent for a RouteTarget constraint
                # route
                routeEvent = RouteEvent(
                    RouteEvent.WITHDRAW,
                    self._subscription2RTCRouteEntry(unsubscription), self)
                log.debug("Based on unsubscription => synthesized withdraw"
                          " for RTC %s", routeEvent)
                self.routeTableManager.enqueue(routeEvent)
            else:
                log.debug("No need to synthesize an RTC route "
                          "(wasLastWorkerForSubscription:%s) ",
                          wasLastWorkerForSubscription)

    # FIXME: this can be subject to races
    def _trackedSubscriptionsAddWorker(self, subs):
        '''returns 1 if this is the first worker subscribed'''

        result = 0
        if (subs.afi, subs.safi, subs.routeTarget) not in self.trackedSubs:
            self.trackedSubs[(subs.afi, subs.safi, subs.routeTarget)] = set()
            result = 1

        self.trackedSubs[
            (subs.afi, subs.safi, subs.routeTarget)].add(subs.worker)

        return result

    # FIXME: this can be subject to races
    def _trackedSubscriptionsRemoveWorker(self, subs):
        '''returns 1 if this was the last worker subscribed'''

        self.trackedSubs[(subs.afi, subs.safi, subs.routeTarget)
                         ].remove(subs.worker)

        if len(self.trackedSubs[(subs.afi, subs.safi, subs.routeTarget)]) == 0:
            del self.trackedSubs[(subs.afi, subs.safi, subs.routeTarget)]
            return 1
        else:
            return 0

    def _subscription2RTCRouteEntry(self, subscription):

        route = Route(RouteTargetConstraint(AFI(AFI.ipv4), SAFI(
            SAFI.rtc), self.config['my_as'], subscription.routeTarget))
        nh = Inet(
            1, socket.inet_pton(socket.AF_INET, self.config['local_address']))
        route.attributes.add(NextHop(nh))

        routeEntry = RouteEntry(AFI(AFI.ipv4), SAFI(
            SAFI.rtc), [], route.nlri, route.attributes, self)

        return routeEntry

    # Looking Glass Functions ###################

    def getLGMap(self):
        return {"peers":   (LGMap.COLLECTION,
                            (self.getLGPeerList, self.getLGPeerPathItem)),
                "routes":  (LGMap.FORWARD, self.routeTableManager),
                "workers": (LGMap.FORWARD, self.routeTableManager), }

    def getEstablishedPeersCount(self):
        return reduce(lambda count, peer: count +
                      (isinstance(peer, BGPPeerWorker) and
                       peer.isEstablished()),
                      self.peers.itervalues(), 0)

    def getLGPeerList(self):
        return [{"id": peer.peerAddress,
                 "state": peer.fsm.state} for peer in self.peers.itervalues()]

    def getLGPeerPathItem(self, pathItem):
        return self.peers[pathItem]



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from netaddr.ip import IPNetwork

from bagpipe.bgp.vpn.dataplane_drivers import VPNInstanceDataplane
from bagpipe.bgp.vpn.dataplane_drivers import DataplaneDriver

from bagpipe.bgp.common.looking_glass import LookingGlass, \
    LookingGlassLocalLogger


from bagpipe.bgp.common import logDecorator


BRIDGE_NAME_PREFIX = "bns"
BRIDGE_INTERFACE_PREFIX = "bn-"
LINUX_DEV_LEN = 14
NAMESPACE_INTERFACE_PREFIX = "ns-"


class MPLSLinuxVRFDataplane(VPNInstanceDataplane, LookingGlass):

    '''
    Dataplane driver using MPLS code at github.com/i-maravic/MPLS-Linux.git
    mpls-new branch (based on Linux 3.7rc4)
    '''

    def __init__(self, *args, **kwargs):
        VPNInstanceDataplane.__init__(self, *args)

        self.namespaceId = self._get_namespace_from_network()

        self.log.info("VRF %d: Initializing network namespace %s",
                      self.instanceId, self.namespaceId)
        if self._namespace_exists():
            self.log.debug("VRF netns already exists, flushing MPLS routes...")
            # Flush all MPLS routes in network namespace
            (output, _) = self._runCommand(
                "ip netns exec %s ip route show" % self.namespaceId)
            for line in output:
                if "mpls" in line:
                    self._runCommand("ip netns exec %s ip route del %s" %
                                     (self.namespaceId, line))
        else:
            self.log.debug("VRF network namespace doesn't exist, creating...")
            # Create network namespace
            self._runCommand("ip netns add %s" % self.namespaceId)

            # Set up mpls0 interface
            self._runCommand(
                "ip netns exec %s ip link set mpls0 up" % self.namespaceId)

            # Set up veth pair devices
            (tap_dev, ns_dev) = self._create_veth_pair()

            # Retrieve broadcast IP address
            ip = IPNetwork("%s/%s" % (self.gatewayIP, self.mask))
            broadcastIP = str(ip.broadcast)

            # Set up bridge network namespace interface as gateway
            self._runCommand("ip netns exec %s ip addr add %s/%s broadcast "
                             "%s dev %s" %
                             (self.namespaceId, self.gatewayIP,
                              self.mask, broadcastIP, ns_dev),
                             raiseExceptionOnError=False)

            # Setup IP forwarding
            self._runCommand("ip netns exec %s sh -c \"echo 1 > /proc/sys/"
                             "net/ipv4/ip_forward\"" % self.namespaceId)
            self._runCommand("ip netns exec %s sh -c \"echo 1 > /proc/sys/net"
                             "/ipv4/conf/all/forwarding\"" % self.namespaceId)

            # Setup ARP proxying
            self._runCommand("ip netns exec %s sh -c \"echo 1 > /proc/sys/net"
                             "/ipv4/conf/%s/proxy_arp\"" %
                             (self.namespaceId, ns_dev))
            self._runCommand("ip netns exec %s sh -c \"echo 1 > /proc/sys/net"
                             "/ipv4/conf/%s/proxy_arp_pvlan\"" %
                             (self.namespaceId, ns_dev))

            # Create bridge and adds tap interface on it
            self._create_namespace_bridge(tap_dev)

    @logDecorator.logInfo
    def cleanup(self):
        bridge_name = self._get_bridge_name()

        self.log.warning("Cleaning routes that redirect traffic to network"
                         " namespace %s  (NOT IMPLEMENTED CORRECTLY YET!)",
                         self.namespaceId)
        # FIXME: bogus!! will in fact delete all MPLS routes, including for
        # other VRFs !!
        (output, _) = self._runCommand("ip -M route show")
        for line in output:
            if "netns" in line:
                self._runCommand("ip -M route del %s" % line)

        self.log.info("Cleaning VRF bridge %s", bridge_name)
        self._runCommand("ip link set %s down", bridge_name)
        self._runCommand("brctl delbr %s", bridge_name)

        self.log.info("Cleaning VRF namespace %s", self.namespaceId)
        self._runCommand("ip netns delete %s" % self.namespaceId)

    def _get_bridge_dev_name(self):
        return (BRIDGE_INTERFACE_PREFIX + self.namespaceId)[:LINUX_DEV_LEN]

    def _get_namespace_from_network(self):
        return self.externalInstanceId[:LINUX_DEV_LEN]

    def _get_ns_dev_name(self):
        return (NAMESPACE_INTERFACE_PREFIX + self.namespaceId)[:LINUX_DEV_LEN]

    def _get_bridge_name(self):
        return (BRIDGE_NAME_PREFIX + self.namespaceId)[:LINUX_DEV_LEN]

    def _namespace_exists(self):
        """ Check if network namespace exist. """
        (output, _) = self._runCommand("ip netns show")
        return (self.namespaceId in output)

    def _create_veth_pair(self):
        """ Create a pair of veth devices """
        bridge_to_ns = self._get_bridge_dev_name()
        ns_to_bridge = self._get_ns_dev_name()

        self._runCommand("ip link add %s type veth peer name %s netns %s" %
                         (bridge_to_ns, ns_to_bridge, self.namespaceId))
        self._runCommand("ip link set dev %s up" % bridge_to_ns)
        self._runCommand("ip netns exec %s ip link set dev %s up" %
                         (self.namespaceId, ns_to_bridge))

        return (bridge_to_ns, ns_to_bridge)

    def _bridge_exists(self, bridge):
        """Check if bridge exists."""
        # TODO: this code generate an ERROR self.log even when this is not a
        #       real error should use acceptableReturnCodes of _runCommand
        try:
            self._runCommand("ip link show dev %s" % bridge)
        except Exception:
            return False
        return True

    def _create_namespace_bridge(self, interface):
        """ Create 'bridge' and add 'interface' to it """
        bridge_name = self._get_bridge_name()

        if not self._bridge_exists(bridge_name):
            self.log.debug("Starting bridge %s", bridge_name)
            self._runCommand("brctl addbr %s" % bridge_name)
            self._runCommand("brctl setfd %s 0" % bridge_name)
            self._runCommand("brctl stp %s off" % bridge_name)
            self._runCommand("ip link set %s up" % bridge_name)

            self._runCommand("brctl addif %s %s" % (bridge_name, interface))
            self.log.debug("Bridge %s started with interface "
                           "%s added", bridge_name, interface)

    @logDecorator.logInfo
    def vifPlugged(self, macAddress, ipAddress, localPort, label):

        if "lo" in localPort:
            self.log.debug("vifPlugged: Plugging loopback interface")
            self._runCommand("ip netns exec %s ip link set %s up" %
                             (self.namespaceId, localPort))
            self._runCommand("ip netns exec %s ip addr add %s/32 dev %s" %
                             (self.namespaceId, ipAddress, localPort),
                             raiseExceptionOnError=False)
        else:
            self.log.debug("vifPlugged: Plugging local port %s", localPort)
            bridge_name = self._get_bridge_name()

            # Attach VIF on network namespace bridge
            self._runCommand("brctl addif %s %s" % (bridge_name, localPort),
                             raiseExceptionOnError=False)

        # Add ip route to redirect traffic to the correct namespace depending
        # on MPLS label
        self._runCommand("ip -M route add %d mpls pop 1 netns %s" %
                         (label, self.namespaceId))

        if self.log.debug:
            self._runCommand("ip -M route show")
            self._runCommand("ip netns exec %s ip route show" %
                             self.namespaceId)

    @logDecorator.logInfo
    def vifUnplugged(self, macAddress, ipAddress, localPort, label,
                     lastEndpoint=True):

        # Remove ip route to disable traffic redirection to namespace depending
        # on MPLS label
        self._runCommand("ip -M route del %d mpls pop 1 netns %s" %
                         (label, self.namespaceId))

        if "lo" in localPort:
            self.log.debug("vifUnplugged: Unplugging loopback interface")
            self._runCommand("ip netns exec %s ip link set %s down" %
                             (self.namespaceId, localPort))
            self._runCommand("ip netns exec %s ip addr del %s/32 dev %s" %
                             (self.namespaceId, ipAddress, localPort))
        else:
            self.log.debug("vifUnplugged: Unplugging local port %s", localPort)
            bridge_name = self._get_bridge_name()

            # Detach VIF from network namespace bridge
            self._runCommand("brctl delif %s %s" %
                             (bridge_name, localPort),
                             raiseExceptionOnError=False)

    @logDecorator.logInfo
    def setupDataplaneForRemoteEndpoint(self, prefix, remotePE, label, nlri,
                                        encaps):
        self._runCommand("ip netns exec %s ip route replace %s mpls push %d"
                         " global dev %s %s" %
                         (self.namespaceId, prefix, label,
                          self.driver.mpls_interface, remotePE))

    def removeDataplaneForRemoteEndpoint(self, prefix, remotePE, label, nlri):
        self.log.info("removeDataplaneForRemoteEndpoint(%s, %s, %d, %s)",
                      prefix, remotePE, label, nlri)

        self._runCommand("ip netns exec %s ip route del %s mpls push %d global"
                         " dev %s %s" % (self.namespaceId, prefix, label,
                                         self.driver.mpls_interface, remotePE))


class MPLSLinuxDataplaneDriver(DataplaneDriver, LookingGlass):

    """
    This dataplane driver relies on the MPLS stack for the Linux kernel at:
         https://github.com/i-maravic/iproute2/tree/mpls-new

    This kernel module is based on a Linux 3.7x version.

    This driver requires the corresponding iproute utility at:
         https://github.com/i-maravic/iproute2/tree/mpls-new

    This driver should be considered **obsolete** as this MPLS stack seems
    unmaintained.
    It wasn't tested against the most recent evolutions of bagpipe-bgp and
    may not fully work.
    """

    dataplaneInstanceClass = MPLSLinuxVRFDataplane

    def __init__(self, config, init=True):
        LookingGlassLocalLogger.__init__(self)
        DataplaneDriver.__init__(self, config, init)

    @logDecorator.logInfo
    def _initReal(self, config):
        self.config = config

        self._runCommand("modprobe mpls")

        if "*gre*" in self.config["mpls_interface"]:
            self.mpls_interface = "gre_wildcard"
            self._runCommand("ip tunnel add %s mode gre local %s remote "
                             "0.0.0.0" %
                             (self.mpls_interface, self.getLocalAddress()),
                             raiseExceptionOnError=False)
            self._runCommand("ip link set %s up" % self.mpls_interface)
            self._runCommand("ip link set %s mpls on" % self.mpls_interface)
        else:
            self.mpls_interface = self.config["mpls_interface"]
            self._runCommand("ip link set %s mpls on" % self.mpls_interface)

        self._runCommand("ip link set mpls0 up")
        self._runCommand("ip link set lo mpls on")

    @logDecorator.logInfo
    def resetState(self):
        # Flush all MPLS routes redirecting traffic to network namespaces
        (output, _) = self._runCommand("ip -M route show")
        for line in output:
            if "netns" in line:
                self._runCommand("ip -M route del %s" % line)

        if self.log.debug:
            self.log.debug("All MPLS routes have been flushed")
            self._runCommand("ip -M route show")

        # Flush all namespaces MPLS routes
        (output, _) = self._runCommand("ip netns show")
        namespace_list = output
        if not namespace_list:
            self.log.debug("No namespaces are configured")
        else:
            for namespaceId in output:
                for line in self._runCommand("ip netns exec %s ip route show" %
                                             namespaceId)[0]:
                    if "mpls" in line:
                        self._runCommand("ip netns exec %s ip route del %s" %
                                         (namespaceId, line))

                if self.log.debug:
                    self.log.debug("Netns %s: MPLS routes have been flushed",
                                   namespaceId)
                    self._runCommand("ip netns exec %s ip route show" %
                                     namespaceId)

    def _cleanupReal(self):
        self.log.warning("not implemented yet!")



# encoding: utf-8
"""
Copyright (c) 2014, Orange
All rights reserved.

File released under the BSD 3-Clause license.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions 
are met:

1. Redistributions of source code must retain the above copyright 
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in 
   the documentation and/or other materials provided with the 
   distribution.

3. Neither the name of the copyright holder nor the names of its 
   contributors may be used to endorse or promote products derived 
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
"""

from struct import pack,unpack
import socket

from bagpipe.exabgp.structure.ip import _Prefix
from bagpipe.exabgp.structure.ip import _bgp as len_to_bytes
from bagpipe.exabgp.structure.ip import Inet
from bagpipe.exabgp.structure.address    import AFI,SAFI

from bagpipe.exabgp.structure.mpls import unpackLabelStack

class RouteDistinguisher(object):

    TYPE_AS2_LOC = 0    # Format AS(2bytes):AN(4bytes) 
    TYPE_IP_LOC  = 1    # Format IP address:AN(2bytes)
    TYPE_AS4_LOC = 2    # Format AS(4bytes):AN(2bytes)

    def __init__(self,rdtype,asn,ip,loc):
        self.type = rdtype

        if rdtype in (self.TYPE_AS2_LOC, self.TYPE_AS4_LOC):
            self.asn = asn
            self.loc = loc 
            self.ip = ""
        elif rdtype == self.TYPE_IP_LOC:
            self.ip = ip
            self.loc = loc
            self.asn = 0
        else: 
            raise Exception("unsupported rd rdtype")
        
    def __str__ (self):
        if self.type in(self.TYPE_AS2_LOC,self.TYPE_AS4_LOC):
            return "%s:%s" % (self.asn, self.loc)
        elif self.type == self.TYPE_IP_LOC:
            return "%s:%s" % (self.ip, self.loc)
        else:
            raise "BROKEN RD / UNKNOWN TYPE"

    def __len__(self):
        return 8

    def __repr__ (self):
        return str(self)
    
    def __cmp__(self,other):
        if (isinstance(other,RouteDistinguisher) 
            and self.type == other.type and
            self.asn == other.asn and
            self.ip == other.ip and
            self.loc == other.loc ):
            return 0
        else:
            return -1
    
    def pack(self):
        if self.type == self.TYPE_AS2_LOC:
            return pack( '!HHL', self.type, self.asn, self.loc)
        elif self.type == self.TYPE_IP_LOC:
            encoded_ip = socket.inet_pton(socket.AF_INET, self.ip )
            return pack( '!H4sH', self.type, encoded_ip, self.loc)
        elif self.type == self.TYPE_AS4_LOC:
            return pack( '!HLH', self.type, self.asn, self.loc)
        else:
            raise Exception("Incorrect RD type %d // not supposed to happen !!" % self.type)
        
    @staticmethod
    def unpack(data):
        rdtype = unpack( '!H', data[0:2] )[0]
        data = data[2:]
    
        if rdtype == RouteDistinguisher.TYPE_AS2_LOC:
            asn,loc = unpack("!HL",data )
            ip = None
        elif rdtype == RouteDistinguisher.TYPE_IP_LOC:
            ip = socket.inet_ntop(socket.AF_INET, data[0:4])
            loc = unpack( '!H', data[4:])[0]
            asn = None
        elif rdtype == RouteDistinguisher.TYPE_AS4_LOC:
            asn,loc = unpack("!LH",data )
            ip = None
        else: 
            raise Exception("unsupported rd rdtype: %d" % rdtype)
        
        return RouteDistinguisher(rdtype,asn,ip,loc)


class VPNLabelledPrefix(object):

    def __init__(self,afi,safi,prefix,rd,labelStack):
        self.afi = AFI(afi)
        self.safi = SAFI(safi)
        self.rd = rd
        self.labelStack = labelStack  # an array of LabelStackEntry's
        if type(labelStack) != list or len(labelStack)==0:
            raise Exception("Labelstack has to be a non-empty array")
        self.prefix = prefix
    
    def __str__ (self):
        return "RD:%s %s MPLS:[%s]" % (self.rd, self.prefix, "|".join(map(str,self.labelStack)))

    def __repr__(self):
        return self.__str__() 

    def pack(self):
        bitlen = (len(self)-len(self.prefix))*8 + self.prefix.mask
        
        stack = ''.join( map(lambda x:x.pack(), self.labelStack ) )
        
        return chr(bitlen) + stack + self.rd.pack() + self.prefix.pack()[1:]
        
    def __len__(self):
        # returns the length in bits!
        return    len(self.labelStack) * len(self.labelStack[0]) + len(self.rd) + len(self.prefix)
        
        
    def __cmp__(self,other):
        #
        # Note well: we need an advertise and a withdraw for the same RD:prefix to result in 
        # objects that are equal for Python, this is why the test below does not look at self.labelstack
        #
        if (isinstance(other,VPNLabelledPrefix) and
            self.rd == other.rd and
            self.prefix == other.prefix):
            return 0
        else:
            return -1
        
    def __hash__(self):  #FIXME: improve for better performance?
        return hash("%s:%s" % (self.rd,self.prefix))
    
    @staticmethod
    def unpack(afi,safi,data):
                
        # prefix len
        bitlen = ord(data[0]) 
        data=data[1:]
        initial_len = len(data)
        
        # data is supposed to be: label stack, rd, prefix
        labelStack,consummed = unpackLabelStack(data)
        data=data[consummed:]
        
        rd = RouteDistinguisher.unpack(data[:8])
        data=data[8:]
        
        prefix_len_in_bits = bitlen - (initial_len - len(data))*8
        last_byte = len_to_bytes[prefix_len_in_bits]
        prefix = _Prefix(afi, data[0:last_byte] + '\0'*(Inet._length[afi]-last_byte) , prefix_len_in_bits )
        
        return VPNLabelledPrefix(afi,safi,prefix,rd,labelStack)



# encoding: utf-8
"""
asn.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""


from struct import pack

# =================================================================== ASN

class ASN (int):
	def asn4 (self):
		return self > pow(2,16)

	def pack (self,asn4):
		if asn4:
			return pack('!L',self)
		return pack('!H',self)

	def __len__ (self):
		if self.asn4:
			return 4
		return 2

AS_TRANS = ASN(23456)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
import time

import urllib

import traceback

import re
import json

from bottle import request, response, abort, Bottle

from bagpipe.bgp.common.looking_glass import LookingGlass, LGMap, \
    NoSuchLookingGlassObject, LookingGlassReferences

log = logging.getLogger(__name__)

LOOKING_GLASS_BASE = "looking-glass"


class APIException(Exception):
    pass


class RESTAPI(LookingGlass):

    """BGP component REST API."""

    # Random generated sequence number
    BGP_SEQ_NUM = int(uuid.uuid4())

    def __init__(self, config, daemon, vpnManager, catchAllLGLogHandler):
        self.config = config
        self.daemon = daemon

        self.vpnManager = vpnManager
        self.catchAllLGLogHandler = catchAllLGLogHandler

        self.bottle = Bottle()

        # Wrapping route callbacks (instead of using decorator) to url
        self.bottle.get("/ping", callback=self.ping)
        self.bottle.post("/attach_localport", callback=self.attach_localport)
        self.bottle.post("/detach_localport", callback=self.detach_localport)

        self.bottle.get("/%s<path:path>" %
                        LOOKING_GLASS_BASE, callback=self.looking_glass)
        self.bottle.get("/%s" %
                        LOOKING_GLASS_BASE, callback=self.looking_glass_root)

        self.bottle.error_handler[500] = self.error500

        self.startTime = time.time()

        LookingGlassReferences.setRoot(LOOKING_GLASS_BASE)
        LookingGlassReferences.setReferencePath(
            "BGP_WORKERS", ["bgp", "workers"])
        LookingGlassReferences.setReferencePath(
            "VPN_INSTANCES", ["vpns", "instances"])

    def ping(self):
        log.debug(
            'Ping received, returning sequence number: %d', self.BGP_SEQ_NUM)
        return "%d" % self.BGP_SEQ_NUM

    def _check_attach_parameters(self, params, attach):
        log.debug("checking params: %s", params)
        paramList = ('vpn_instance_id', 'mac_address', 'ip_address',
                     'local_port')
        if attach:
            paramList += ('vpn_type', 'import_rt', 'export_rt', 'gateway_ip')

        for paramName in paramList:
            if paramName not in params:
                log.warning("Mandatory parameter '%s' is missing", paramName)
                abort(400, "Mandatory parameter '%s' is missing" % paramName)

        # if local_port is not a dict, then assume it designates a linux
        # interface
        if (isinstance(params['local_port'], str)
                or isinstance(params['local_port'], unicode)):
            params['local_port'] = {'linuxif': params['local_port']}

        # if import_rt or export_rt are strings, convert them into lists
        for param in ('import_rt', 'export_rt'):
            if isinstance(params[param], str) or isinstance(params[param],
                                                            unicode):
                try:
                    params[param] = re.split(',+ *', params[param])
                except:
                    abort(400, "Unable to parse string into a list: '%s'" %
                          params[param])

        if not ('linuxif' in params['local_port'] or
                'evpn' in params['local_port']):
            abort(400, "Mandatory key is missing in local_port parameter"
                  "(linuxif, or evpn)")

        if not isinstance(params.get('advertise_subnet', False), bool):
            abort(400, "'advertise_subnet' must be a boolean")

        return params

    def attach_localport(self):
        """
        'vpn_instance_id: external VPN instance identifier (all ports with same
                         vpn_instance_id will be plugged in the same VPN
                         instance
        'vpn_type': type of the VPN instance ('ipvpn' or 'evpn')
        'import_rt': list of import Route Targets (or comma-separated string)
        'export_rt': list of export Route Targets (or comma-separated string)
        'gateway_ip': IP address of gateway for this VPN instance
        'mac_address': MAC address of endpoint to connect to the VPN instance
        'ip_address': IP/mask of endpoint to connect to the VPN instance
        'advertise_subnet': optional, if set to True then VRF will advertise
                            the whole subnet (defaults to False, readvertise
                            ip_address as a singleton (/32)
        'linuxbr': Name of a linux bridge to which the linuxif is already
                 plugged-in (optional)
        'local_port': local port to plug to the VPN instance
            should be a dict containing any of the following key,value pairs
            {
                'linuxif': 'tap456abc', # name of a linux interface
                                        # - if OVS information is provided it
                                        #   does not have to be an existing
                                        #    interface
                                        # -  not needed/not used if 'evpn' plug
                                        #    is used
                'ovs': {                     # optional
                        # whether or not interface is already plugged into the
                        # OVS bridge:
                        'plugged': True,
                        # name of a linux interface to be plugged into the OVS
                        # bridge (optional and ignored if port_number is
                        # provided):
                        'port_name': 'qvo456abc',
                        # OVS port number (optional if 'port_name' provided):
                        'port_number': '7',
                        # the VLAN id for VM traffic (optional)
                        'vlan': '42',
                        # optional specification of a distinct port to send
                        # traffic to the VM(only applies if a vlan is
                        # specified) :
                        'to_vm_port_number'
                        'to_vm_port_name'
                       },
                'evpn': {  # for an ipvpn attachement...
                     'id': 'xyz'  # specifies the vpn_instance_id of an evpn
                                  # that will be attached to the ipvpn
                     'ovs_port_name': 'qvb456abc' # optional, if provided,
                                                  # and if ovs/port_name is
                                                  # also provided, then the
                                                  # interface name will be
                                                  # assumed as already plugged
                                                  # into the evpn
                    }
            }
            if local_port is not a list, it is assumed to be a name of a linux
            interface (string)
        'readvertise': {  # optional, used to re-advertise addresses...
            'from_rt': [list of RTs]  # ...received on these RTs
            'to_rt': [list of RTs] # ...toward these RTs
        }
        'fallback': # (optional) if provided, on a VRF lookup miss,
                    # the MAC destination address will be
                    # rewritten to this MAC before being
                    # sent back where it came from
                    {
                    'src_mac': 'aa:bb:cc:dd:ee:ff'  # new source MAC
                    'dst_mac': 'aa:bb:cc:dd:ee:00'  # new destination MAC
                    'ovs_port_name': 'patch_foo'
                    'ovs_port_number': 4               # (unsupported yet)
                    'ovs_resubmit': '(<port>,<table>)' # (unsupported yet)
        }
        """

        try:
            attach_params = request.json
        except Exception:
            log.error('attach_localport: No local port details received')
            abort(400, 'No local port details received')

        attach_params = self._check_attach_parameters(attach_params, True)

        try:
            log.debug('Local port attach received: %s', attach_params)
            self.vpnManager.plugVifToVPN(attach_params['vpn_instance_id'],
                                         attach_params['vpn_type'],
                                         attach_params['import_rt'],
                                         attach_params['export_rt'],
                                         attach_params['mac_address'],
                                         attach_params['ip_address'],
                                         attach_params['gateway_ip'],
                                         attach_params['local_port'],
                                         attach_params.get('linuxbr'),
                                         attach_params.get('advertise_subnet',
                                                           False),
                                         attach_params.get('readvertise'),
                                         attach_params.get('fallback'))
        except APIException as e:
            log.warning('attach_localport: API parameter error: %s', e)
            abort(400, "API parameter error: %s" % e)
        except Exception as e:
            log.error('attach_localport: An error occurred during local port'
                      ' plug to VPN: %s', e)
            log.info(traceback.format_exc())
            abort(500, 'An error occurred during local port plug to VPN')

    def detach_localport(self):

        try:
            detach_params = request.json
        except Exception:
            log.error('detach_localport: No local port details received')
            abort(400, 'No local port details received')

        detach_params = self._check_attach_parameters(detach_params, False)

        try:
            log.debug('Local port detach received: %s', detach_params)
            self.vpnManager.unplugVifFromVPN(detach_params['vpn_instance_id'],
                                             detach_params['mac_address'],
                                             detach_params['ip_address'],
                                             detach_params['local_port'],
                                             detach_params.get(
                                                 'advertise_subnet', False),
                                             )
        except APIException as e:
            log.warning('detach_localport: API parameter error: %s', e)
            abort(400, "API parameter error: %s" % e)
        except Exception as e:
            log.error('detach_localport: An error occurred during local port'
                      ' unplug from VPN: %s', e)
            log.info(traceback.format_exc())
            abort(500, 'An error occurred during local port unplug from VPN')

    def looking_glass_root(self):
        return self.looking_glass('/')

    def looking_glass(self, path):
        urlPathElements = [urllib.unquote(elem)
                           for elem in path.split('/') if elem is not '']

        pathPrefix = "%s://%s/%s" % (
            request.environ['wsgi.url_scheme'],  # http
            request.environ['HTTP_HOST'],
            LOOKING_GLASS_BASE
        )
        log.debug("pathPrefix: %s", pathPrefix)
        log.debug("urlPathElements: %s", urlPathElements)

        try:
            lgInfo = self.getLookingGlassInfo(pathPrefix, urlPathElements)

            log.debug("lgInfo: %s...", repr(lgInfo)[:40])

            if lgInfo is None:
                raise NoSuchLookingGlassObject(pathPrefix, urlPathElements[0])

            response.content_type = 'application/json'
            return json.dumps(lgInfo)
        except NoSuchLookingGlassObject as e:
            log.info('looking_glass: %s', repr(e))
            abort(404, repr(e))
        except Exception as e:
            log.error('looking_glass: An error occurred: %s', e)
            log.error(traceback.format_exc())
            abort(500, 'Server error')

    # Looking glass hooks #################

    def getLGMap(self):
        return {
            "summary":  (LGMap.SUBITEM, self.getLGSummary),
            "config":   (LGMap.DELEGATE, self.daemon),
            "bgp":      (LGMap.DELEGATE, self.vpnManager.bgpManager),
            "vpns":     (LGMap.DELEGATE, self.vpnManager),
            "logs":     (LGMap.SUBTREE, self.getLogs),
        }

    def getLGSummary(self):
        return {
            "BGP_established_peers":
                self.vpnManager.bgpManager.getEstablishedPeersCount(),
            "local_routes_count":
                self.vpnManager.bgpManager.routeTableManager.
                getLocalRoutesCount(),
            "received_routes_count":
                self.vpnManager.bgpManager.routeTableManager.
                getReceivedRoutesCount(),
            "vpn_instances_count": self.vpnManager.getVPNInstancesCount(),
            "warnings_and_errors": len(self.catchAllLGLogHandler),
            "start_time": time.strftime("%Y-%m-%d %H:%M:%S",
                                        time.localtime(self.startTime))
        }

    def getLogs(self, pathPrefix):
        return [{'level': record.levelname,
                 'time':
                 self.catchAllLGLogHandler.formatter.formatTime(record),
                 'name': record.name,
                 'message': record.msg}
                for record in self.catchAllLGLogHandler.getRecords()]

    def error500(self, error):
        log.error("Bottle catched an error: %s", error.exception)
        log.error(traceback.format_exc())
        return "Server Error"

    def run(self):
        # TODO: make looking-glass available to remote hosts
        self.bottle.run(host=self.config.get("api_host", "localhost"),
                        port=self.config.get("api_port", 8082),
                        quiet=True,
                        debug=True)



# encoding: utf-8
"""
aspath.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

from bagpipe.exabgp.message.update.attribute import AttributeID,Flag,Attribute

# =================================================================== ASPath (2)

class ASPath (Attribute):
	AS_SET      = 0x01
	AS_SEQUENCE = 0x02

	ID = AttributeID.AS_PATH
	FLAG = Flag.TRANSITIVE
	MULTIPLE = False

	def __init__ (self,asn4=False,asptype=0x02,aspsegment=None):
		self.asn4 = asn4
		self.asptype = asptype
		if aspsegment == None:
			self.aspsegment = []
		else:
			self.aspsegment = aspsegment

	def _segment (self,seg_type,values):
		l = len(values)
		if l:
			if l>255:
				return self._segment(seg_type,values[:255]) + self._segment(seg_type,values[255:])
			return "%s%s%s" % (chr(seg_type),chr(len(values)),''.join([v.pack(self.asn4) for v in values]))
		return ""

	def add (self,asn):
		self.aspsegment.append(asn)

	def pack (self):
		return self._attribute(self._segment(self.asptype,self.aspsegment))

	def __len__ (self):
		return 2 + (len(self.aspsegment)*2)

	def __str__ (self):
		if self.asptype == 0x01: t = 'AS_SET'
		if self.asptype == 0x02: t = 'AS_SEQUENCE'
		else: t = 'INVALID'

		if len(self) >  1: return '%s [ %s ]' % (t,' '.join([str(community) for community in self.aspsegment]))
		if len(self) == 1: return '%s %s' % (t,str(self.aspsegment[0]))
		return t

	def __repr__ (self):
		return str(self)
	
	def __cmp__(self,other):
		if (isinstance(other,ASPath) and
			self.asn4 == other.asn4 and
			self.asptype == other.asptype and
			self.aspsegment == other.aspsegment 
		    ):
			return 0
		else:
			return -1

class AS4Path (ASPath):
	ID = AttributeID.AS4_PATH
	FLAG = Flag.TRANSITIVE|Flag.OPTIONAL

	def __init__ (self,asptype=0x02,aspsegment=None):
		ASPath.__init__(self,True,asptype,aspsegment)



# encoding: utf-8
"""
configuration.py

Created by Thomas Mangin on 2011-03-29.
Copyright (c) 2011-2012 Exa Networks. All rights reserved.
"""

import os
import sys

debug = os.environ.get('PDB',None)

def bug_report (type, value, trace):
	import traceback
	from bagpipe.exabgp.log import Logger
	logger = Logger()

	print
	print
	print "-"*80
	print "-- Please provide the information below on :"
	print "-- http://code.google.com/p/exaproxy/issues/entry"
	print "-"*80
	print
	print
	print '-- Version'
	print
	print
	print sys.version
	print
	print
	print "-- Configuration"
	print
	print
	print logger.config()
	print
	print
	print "-- Logging History"
	print
	print
	print logger.history()
	print
	print
	print "-- Traceback"
	print
	print
	traceback.print_exception(type,value,trace)
	print
	print
	print "-"*80
	print "-- Please provide the information above on :"
	print "-- http://code.google.com/p/exaproxy/issues/entry"
	print "-"*80
	print
	print

	#print >> sys.stderr, 'the program failed with message :', value

if debug is None:
	def intercept_nopdb (type, value, trace):
		bug_report(type, value, trace)
	sys.excepthook = intercept_nopdb
elif debug not in ['0','']:
	def intercept_pdb (type, value, trace):
		bug_report(type, value, trace)
		import pdb
		pdb.pm()
	sys.excepthook = intercept_pdb

del sys.argv[0]

if sys.argv:
	__file__ = os.path.abspath(sys.argv[0])
	__name__ = '__main__'
	execfile(sys.argv[0])



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

.. module:: test_vpn_instance
   :synopsis: module that defines several test cases for the vpn_instance
              module.
   In particular, unit tests for VPNInstance class.
   Setup : Start VPNInstance thread instance.
   TearDown : Stop VPNInstance thread instance.
   VPNInstance is a base class for objects used to manage an E-VPN instance
   (EVI) or IP-VPN instance (VRF)
   Tests are organized as follow :
   - testAx use cases to test endpoints plug with different combinations of MAC
     and IP addresses on a port
   - testBx use cases to test enpoints plug with different combinations of MAC
     and IP addresses on different ports
   - testCx use cases to test endpoints unplug with different combinations of
     MAC and IP addresses as the one plugged on a port
   - testDx use cases to test endpoints unplug with different combinations of
     MAC and IP addresses as the ones plugged on different ports

"""
import mock

from testtools import TestCase
from bagpipe.bgp.tests import RT1, NLRI1

from bagpipe.bgp.vpn.label_allocator import LabelAllocator
from bagpipe.bgp.vpn.vpn_instance import VPNInstance

from bagpipe.exabgp.message.update.attributes import Attributes

from bagpipe.bgp.engine import RouteEntry

from bagpipe.exabgp.structure.address import AFI, SAFI

MAC1 = "00:00:de:ad:be:ef"
IP1 = "10.0.0.2/32"
LOCAL_PORT1 = {'linuxif': 'tap1'}

MAC2 = "00:00:fe:ed:fa:ce"
IP2 = "10.0.0.3/32"
LOCAL_PORT2 = {'linuxif': 'tap2'}

MAC3 = "00:00:de:ad:c0:de"
IP3 = "10.0.0.4/32"
LOCAL_PORT3 = {'linuxif': 'tap3'}

MAC4 = "00:00:fe:ed:f0:0d"
IP4 = "10.0.0.5/32"
LOCAL_PORT4 = {'linuxif': 'tap4'}


class TestableVPNInstance(VPNInstance):

    def _bestRouteRemoved(self, entry, route):
        pass

    def _newBestRoute(self, entry, route, last):
        pass

    def _route2trackedEntry(self, route):
        pass

    def generateVifBGPRoute(self):
        pass


class TestVPNInstance(TestCase):

    def setUp(self):
        super(TestVPNInstance, self).setUp()
        self.labelAllocator = LabelAllocator()

        self.mockDataplane = mock.Mock()
        self.mockDataplane.vifPlugged = mock.Mock()
        self.mockDataplane.vifUnplugged = mock.Mock()

        self.mockDPDriver = mock.Mock()
        self.mockDPDriver.initializeDataplaneInstance.returnValue = \
            self.mockDataplane

        VPNInstance.afi = AFI(AFI.ipv4)
        VPNInstance.safi = SAFI(SAFI.mpls_vpn)
        self.vpnInstance = TestableVPNInstance(mock.Mock(name='BGPManager'),
                                               self.labelAllocator,
                                               self.mockDPDriver, 1, 1,
                                               [RT1], [RT1], '10.0.0.1', 24,
                                               None)
        self.vpnInstance.synthesizeVifBGPRoute = mock.Mock(
            return_value=RouteEntry(self.vpnInstance.afi,
                                    self.vpnInstance.safi, NLRI1, RT1,
                                    Attributes(), None))
        self.vpnInstance._pushEvent = mock.Mock()
        self.vpnInstance._postFirstPlug = mock.Mock()
        self.vpnInstance.start()

    def tearDown(self):
        super(TestVPNInstance, self).tearDown()
        with mock.patch.object(self.vpnInstance.bgpManager, 'cleanup'):
            self.vpnInstance.stop()
            self.vpnInstance.join()

    def _get_ipAddress(self, ipAddressPrefix):
        return ipAddressPrefix[0:ipAddressPrefix.find('/')]

    def _validate_ipAddress2MacAddress_consistency(self, macAddress,
                                                   ipAddress1,
                                                   ipAddress2=None):
        # Validate IP address -> MAC address consistency
        self.assertIn(ipAddress1, self.vpnInstance.ipAddress2MacAddress)

        if ipAddress2:
            self.assertIn(ipAddress1, self.vpnInstance.ipAddress2MacAddress)
            self.assertEquals(
                self.vpnInstance.ipAddress2MacAddress[ipAddress1],
                self.vpnInstance.ipAddress2MacAddress[ipAddress2])
        else:
            self.assertEquals(
                macAddress, self.vpnInstance.ipAddress2MacAddress[ipAddress1])

    def _validate_macAddress2LocalPortData_consistency(self, macAddress,
                                                       localPort):
        # Validate MAC address -> Port informations consistency
        self.assertIn(macAddress, self.vpnInstance.macAddress2LocalPortData)

        port_info = self.vpnInstance.macAddress2LocalPortData[
            macAddress]['port_info']
        self.assertEquals(localPort['linuxif'], port_info['linuxif'])

    def _validate_localPort2Endpoints_consistency(self, length, localPort,
                                                  endpoints):
        # Validate Port -> Endpoint (MAC, IP) tuple consistency
        self.assertEqual(
            length,
            len(self.vpnInstance.localPort2Endpoints[localPort['linuxif']]))

        for macAddress, ipAddress in endpoints:
            endpoint_info = {'mac': macAddress, 'ip': ipAddress}
            self.assertIn(
                endpoint_info,
                self.vpnInstance.localPort2Endpoints[localPort['linuxif']])

    def testA1_plugEnpointTwiceSamePort(self):
        '''
        Plug one endpoint with same MAC and IP addresses twice on a port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)

        self.assertEqual(1, self.vpnInstance.dataplane.vifPlugged.call_count,
                         "Port must be plugged only once on dataplane")
        self.assertEqual(1, self.vpnInstance._pushEvent.call_count,
                         "Route for port must be advertised only once")

        self._validate_ipAddress2MacAddress_consistency(MAC1, IP1)
        self._validate_macAddress2LocalPortData_consistency(MAC1, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])

    def testA2_plugMultipleEnpointsWithSameIPSamePort(self):
        '''
        Plug multiple enpoints with different MAC addresses and same IP
        address on a port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)

        # An IP address correspond to only one MAC address, exception must be
        # raised
        self.assertRaises(Exception,
                          self.vpnInstance.vifPlugged,
                          MAC2, IP1, LOCAL_PORT1)
        self.assertEqual(1, self.vpnInstance.dataplane.vifPlugged.call_count,
                         "Only first port must be plugged on dataplane")
        self.assertEqual(1, self.vpnInstance._pushEvent.call_count,
                         "Only route for first port must be advertised")

        self._validate_ipAddress2MacAddress_consistency(MAC1, IP1)
        self._validate_macAddress2LocalPortData_consistency(MAC1, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])
        self.assertNotIn(MAC2, self.vpnInstance.macAddress2LocalPortData)

    def testA3_plugMultipleEndpointsWithSameMACSamePort(self):
        '''
        Plug multiple endpoints with same MAC address and different IP
        addresses on a port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC1, IP2, LOCAL_PORT1)

        self.assertEqual(2, self.vpnInstance.dataplane.vifPlugged.call_count,
                         "Port different IP addresses must be plugged on "
                         "dataplane")
        self.assertEqual(2, self.vpnInstance._pushEvent.call_count,
                         "Route for port different IP addresses must be "
                         "advertised")
        self._validate_ipAddress2MacAddress_consistency(MAC1, IP1, IP2)
        self._validate_macAddress2LocalPortData_consistency(MAC1, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            2, LOCAL_PORT1, [(MAC1, IP1), (MAC1, IP2)])

    def testA4_plugMultipleEndpointsSamePort(self):
        '''
        Plug multiple endpoints with different MAC and IP addresses on a port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC2, IP2, LOCAL_PORT1)

        self.assertEqual(2, self.vpnInstance.dataplane.vifPlugged.call_count,
                         "Port different endpoints must be plugged on "
                         "dataplane")
        self.assertEqual(2, self.vpnInstance._pushEvent.call_count,
                         "Route for port different endpoints must be "
                         "advertised")
        self._validate_ipAddress2MacAddress_consistency(MAC1, IP1)
        self._validate_ipAddress2MacAddress_consistency(MAC2, IP2)
        self._validate_macAddress2LocalPortData_consistency(MAC1, LOCAL_PORT1)
        self._validate_macAddress2LocalPortData_consistency(MAC2, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            2, LOCAL_PORT1, [(MAC1, IP1), (MAC2, IP2)])

    def testB1_plugEndpointTwiceDifferentPort(self):
        '''
        Plug one endpoint with same MAC and IP addresses twice on different
        ports
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)

        # A port correspond to only one MAC address, exception must be raised
        self.assertRaises(Exception,
                          self.vpnInstance.vifPlugged,
                          MAC1, IP1, LOCAL_PORT2)
        self.assertEqual(1, self.vpnInstance.dataplane.vifPlugged.call_count,
                         "Only first port must be plugged on dataplane")
        self.assertEqual(1, self.vpnInstance._pushEvent.call_count,
                         "Only route for first port must be advertised")

        self._validate_ipAddress2MacAddress_consistency(MAC1, IP1)
        self._validate_macAddress2LocalPortData_consistency(MAC1, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])
        self.assertNotIn(
            LOCAL_PORT2['linuxif'], self.vpnInstance.localPort2Endpoints)

    def testB2_plugMultipleEndpointsWithSameIPDifferentPort(self):
        '''
        Plug multiple endpoints with different MAC addresses and same IP
        address on different port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)

        # An IP address correspond to only one MAC address, exception must be
        # raised
        self.assertRaises(Exception,
                          self.vpnInstance.vifPlugged,
                          MAC2, IP1, LOCAL_PORT2)
        self.assertEqual(1, self.vpnInstance.dataplane.vifPlugged.call_count,
                         "Only first port must be plugged on dataplane")
        self.assertEqual(1, self.vpnInstance._pushEvent.call_count,
                         "Only route for first port must be advertised")

        self._validate_ipAddress2MacAddress_consistency(MAC1, IP1)
        self._validate_macAddress2LocalPortData_consistency(MAC1, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])
        self.assertNotIn(
            LOCAL_PORT2['linuxif'], self.vpnInstance.localPort2Endpoints)

    def testB4_plugMultipleEndpointsWithSameMACDifferentPort(self):
        '''
        Plug multiple endpoints with same MAC address and different IP
        addresses on different ports
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)

        # A port correspond to only one MAC address, exception must be raised
        self.assertRaises(Exception,
                          self.vpnInstance.vifPlugged,
                          MAC1, IP2, LOCAL_PORT2)
        self.assertEqual(1, self.vpnInstance.dataplane.vifPlugged.call_count,
                         "Only first port must be plugged on dataplane")
        self.assertEqual(1, self.vpnInstance._pushEvent.call_count,
                         "Only route for first port must be advertised")

        self._validate_ipAddress2MacAddress_consistency(MAC1, IP1)
        self._validate_macAddress2LocalPortData_consistency(MAC1, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])
        self.assertNotIn(
            LOCAL_PORT2['linuxif'], self.vpnInstance.localPort2Endpoints)

    def testB5_plugMultipleEndpointsDifferentPort(self):
        '''
        Plug multiple endpoints with different MAC and IP addresses on
        different ports
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC2, IP2, LOCAL_PORT2)

        self.assertEqual(2, self.vpnInstance.dataplane.vifPlugged.call_count,
                         "All ports must be plugged on dataplane")
        self.assertEqual(2, self.vpnInstance._pushEvent.call_count,
                         "Routes for all ports must be advertised")

        self._validate_ipAddress2MacAddress_consistency(MAC1, IP1)
        self._validate_macAddress2LocalPortData_consistency(MAC1, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])

        self._validate_ipAddress2MacAddress_consistency(MAC2, IP2)
        self._validate_macAddress2LocalPortData_consistency(MAC2, LOCAL_PORT2)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT2, [(MAC2, IP2)])

    def testC1_unplugUniqueEndpointSamePort(self):
        '''
        Unplug one endpoint with same MAC and IP addresses as the one plugged
        on port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)

        label1 = self.vpnInstance.macAddress2LocalPortData[MAC1]['label']

        self.vpnInstance.vifUnplugged(MAC1, IP1)

        self.assertEqual(1, self.vpnInstance.dataplane.vifUnplugged.call_count,
                         "Endpoint could be unplugged from dataplane")
        self.assertEqual(
            [((MAC1, self._get_ipAddress(IP1), LOCAL_PORT1, label1, True),)],
            self.vpnInstance.dataplane.vifUnplugged.call_args_list)
        self.assertEqual(2, self.vpnInstance._pushEvent.call_count,
                         "Route must be first advertised and after withdrawn")

        self.assertEqual({}, self.vpnInstance.macAddress2LocalPortData)
        self.assertEqual({}, self.vpnInstance.ipAddress2MacAddress)
        self.assertEqual({}, self.vpnInstance.localPort2Endpoints)

    def testC2_unplugUniqueEndpointWithSameIPSamePort(self):
        '''
        Unplug one endpoint with different MAC addresses and same IP address as
        the one plugged on port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)

        self.assertRaises(Exception,
                          self.vpnInstance.vifUnplugged,
                          MAC2, IP1)

        self.assertEqual(0, self.vpnInstance.dataplane.vifUnplugged.call_count,
                         "Endpoint could not be unplugged from dataplane")
        self.assertEqual(1, self.vpnInstance._pushEvent.call_count,
                         "Route must only be advertised")

        self.assertIn(MAC1, self.vpnInstance.macAddress2LocalPortData)
        self.assertIn(IP1, self.vpnInstance.ipAddress2MacAddress)
        self.assertIn(
            LOCAL_PORT1['linuxif'], self.vpnInstance.localPort2Endpoints)

    def testC3_unplugUniqueEndpointWithSameMACSamePort(self):
        '''
        Unplug one endpoint with same MAC address and different IP addresses
        as the one plugged on port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)

        self.assertRaises(Exception,
                          self.vpnInstance.vifUnplugged,
                          MAC1, IP2)

        self.assertEqual(0, self.vpnInstance.dataplane.vifUnplugged.call_count,
                         "Endpoint could not be unplugged from dataplane")
        self.assertEqual(1, self.vpnInstance._pushEvent.call_count,
                         "Route must only be advertised")

        self.assertIn(MAC1, self.vpnInstance.macAddress2LocalPortData)
        self.assertIn(IP1, self.vpnInstance.ipAddress2MacAddress)
        self.assertIn(
            LOCAL_PORT1['linuxif'], self.vpnInstance.localPort2Endpoints)

    def testC4_unplugOneEndpointSamePort(self):
        '''
        Unplug only one endpoint with same MAC and IP addresses
        corresponding to one plugged on port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC2, IP2, LOCAL_PORT1)

        label1 = self.vpnInstance.macAddress2LocalPortData[MAC1]['label']

        self.vpnInstance.vifUnplugged(MAC1, IP1)

        self.assertEqual(1, self.vpnInstance.dataplane.vifUnplugged.call_count,
                         "Endpoint must be unplugged from dataplane")
        self.assertEqual(
            [((MAC1, self._get_ipAddress(IP1), LOCAL_PORT1, label1, False),)],
            self.vpnInstance.dataplane.vifUnplugged.call_args_list)
        self.assertEqual(3, self.vpnInstance._pushEvent.call_count,
                         "Routes for all port endpoints must be first "
                         "advertised and only one withdrawn")

        self._validate_ipAddress2MacAddress_consistency(MAC2, IP2)
        self._validate_macAddress2LocalPortData_consistency(MAC2, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT1, [(MAC2, IP2)])

    def testC5_unplugAllEndpointsSamePort(self):
        '''
        Unplug all endpoints with same MAC and IP addresses
        corresponding to those plugged on port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC2, IP2, LOCAL_PORT1)

        label1 = self.vpnInstance.macAddress2LocalPortData[MAC1]['label']
        label2 = self.vpnInstance.macAddress2LocalPortData[MAC2]['label']

        self.vpnInstance.vifUnplugged(MAC1, IP1)
        self.vpnInstance.vifUnplugged(MAC2, IP2)

        self.assertEqual(2, self.vpnInstance.dataplane.vifUnplugged.call_count,
                         "All port endpoints must be unplugged from dataplane")
        self.assertEqual(
            [((MAC1, self._get_ipAddress(IP1), LOCAL_PORT1, label1, False),),
             ((MAC2, self._get_ipAddress(IP2), LOCAL_PORT1, label2, True),)],
            self.vpnInstance.dataplane.vifUnplugged.call_args_list)
        self.assertEqual(4, self.vpnInstance._pushEvent.call_count,
                         "Routes for all port endpoints must be first "
                         "advertised and after withdrawn")

        self.assertEqual({}, self.vpnInstance.macAddress2LocalPortData)
        self.assertEqual({}, self.vpnInstance.ipAddress2MacAddress)
        self.assertEqual({}, self.vpnInstance.localPort2Endpoints)

    def testD1_unplugUniqueEndpointsDifferentPort(self):
        '''
        Unplug the endpoints with different MAC and IP addresses corresponding
        to those plugged on different ports
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC2, IP2, LOCAL_PORT2)

        label1 = self.vpnInstance.macAddress2LocalPortData[MAC1]['label']
        label2 = self.vpnInstance.macAddress2LocalPortData[MAC2]['label']

        self.vpnInstance.vifUnplugged(MAC1, IP1)
        self.vpnInstance.vifUnplugged(MAC2, IP2)

        self.assertEqual(2, self.vpnInstance.dataplane.vifUnplugged.call_count,
                         "All different ports endpoints must be unplugged "
                         "from dataplane")
        self.assertEqual(
            [((MAC1, self._get_ipAddress(IP1), LOCAL_PORT1, label1, True),),
             ((MAC2, self._get_ipAddress(IP2), LOCAL_PORT2, label2, True),)],
            self.vpnInstance.dataplane.vifUnplugged.call_args_list)
        self.assertEqual(4, self.vpnInstance._pushEvent.call_count,
                         "Routes for all different ports endpoints must be "
                         "first advertised and after withdrawn")

        self.assertEqual({}, self.vpnInstance.macAddress2LocalPortData)
        self.assertEqual({}, self.vpnInstance.ipAddress2MacAddress)
        self.assertEqual({}, self.vpnInstance.localPort2Endpoints)

    def testD2_unplugOneEndpointSameIPDifferentPort(self):
        '''
        Unplug one endpoint with different MAC or IP address corresponding to
        one plugged on another port
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC2, IP2, LOCAL_PORT2)

        self.assertRaises(Exception,
                          self.vpnInstance.vifUnplugged,
                          MAC1, IP2)

        self.assertEqual(0, self.vpnInstance.dataplane.vifUnplugged.call_count,
                         "Endpoint could not be unplugged from dataplane")
        self.assertEqual(2, self.vpnInstance._pushEvent.call_count,
                         "Routes for all different ports endpoints must only "
                         "be advertised")

        self._validate_ipAddress2MacAddress_consistency(MAC1, IP1)
        self._validate_macAddress2LocalPortData_consistency(MAC1, LOCAL_PORT1)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT1, [(MAC1, IP1)])

        self._validate_ipAddress2MacAddress_consistency(MAC2, IP2)
        self._validate_macAddress2LocalPortData_consistency(MAC2, LOCAL_PORT2)
        self._validate_localPort2Endpoints_consistency(
            1, LOCAL_PORT2, [(MAC2, IP2)])

    def testD3_unplugMultipleEndpointsDifferentPort(self):
        '''
        Unplug multiple endpoints with same MAC and IP addresses corresponding
        to those plugged on different ports
        '''
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC2, IP2, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC3, IP3, LOCAL_PORT2)
        self.vpnInstance.vifPlugged(MAC4, IP4, LOCAL_PORT2)

        label1 = self.vpnInstance.macAddress2LocalPortData[MAC1]['label']
        label2 = self.vpnInstance.macAddress2LocalPortData[MAC2]['label']
        label3 = self.vpnInstance.macAddress2LocalPortData[MAC3]['label']
        label4 = self.vpnInstance.macAddress2LocalPortData[MAC4]['label']

        self.vpnInstance.vifUnplugged(MAC1, IP1)
        self.vpnInstance.vifUnplugged(MAC2, IP2)
        self.vpnInstance.vifUnplugged(MAC3, IP3)
        self.vpnInstance.vifUnplugged(MAC4, IP4)

        self.assertEqual(4, self.vpnInstance.dataplane.vifUnplugged.call_count,
                         "All different ports endpoints must be unplugged "
                         "from dataplane")
        self.assertEqual(
            [((MAC1, self._get_ipAddress(IP1), LOCAL_PORT1, label1, False),),
             ((MAC2, self._get_ipAddress(
               IP2), LOCAL_PORT1, label2, True),),
             ((MAC3, self._get_ipAddress(
               IP3), LOCAL_PORT2, label3, False),),
             ((MAC4, self._get_ipAddress(IP4), LOCAL_PORT2, label4, True),)],
            self.vpnInstance.dataplane.vifUnplugged.call_args_list)
        self.assertEqual(8, self.vpnInstance._pushEvent.call_count,
                         "Routes for all different ports endpoints must be "
                         "first advertised and after withdrawn")

        self.assertEqual({}, self.vpnInstance.macAddress2LocalPortData)
        self.assertEqual({}, self.vpnInstance.ipAddress2MacAddress)
        self.assertEqual({}, self.vpnInstance.localPort2Endpoints)

    def test_getLGLocalPortData(self):
        self.vpnInstance.vifPlugged(MAC1, IP1, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC2, IP2, LOCAL_PORT1)
        self.vpnInstance.vifPlugged(MAC3, IP3, LOCAL_PORT2)
        self.vpnInstance.vifPlugged(MAC4, IP4, LOCAL_PORT2)

        print "\n"
        print self.vpnInstance.getLGLocalPortData("")



# encoding: utf-8
"""
eor.py

Created by Thomas Mangin on 2010-01-16.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""

from bagpipe.exabgp.structure.address import Address,AFI,SAFI
from bagpipe.exabgp.message.update import Update
from bagpipe.exabgp.message.update.attributes import Attributes

# =================================================================== End-Of-Record

class Empty (Address):
	def __init__ (self,afi,safi):
		Address.__init__(self,AFI(afi),SAFI(safi))

	def pack (self):
		return ''
	def __len__ (self):
		return 0

class EmptyRoute (object):
	autocomplete = False

	def __init__ (self,afi,safi):
		self.attributes = Attributes()
		self.nlri = Empty(afi,safi)

class EOR (object):
	def __init__ (self):
		self._announced = []

	def eors (self,families):
		self._announced = []
		r = []
		for afi,safi in families:
			if afi == AFI.ipv4 and safi in [SAFI.unicast, SAFI.multicast]:
				r.append(self.ipv4())
			else:
				r.append(self.mp(afi,safi))
			self._announced.append((afi,safi))
		return r

	def ipv4 (self):
		return Update([EmptyRoute(AFI.ipv4,SAFI.unicast),]).withdraw()

	def mp (self,afi,safi):
		return Update([EmptyRoute(afi,safi),]).withdraw()

	def announced (self):
		return self._announced



# encoding: utf-8
"""
id.py

Created by Thomas Mangin on 2010-01-19.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""

class AttributeID (int):
	# This should move within the classes and not be here
	# RFC 4271
	ORIGIN             = 0x01
	AS_PATH            = 0x02
	NEXT_HOP           = 0x03
	MED                = 0x04
	LOCAL_PREF         = 0x05
	ATOMIC_AGGREGATE   = 0x06
	AGGREGATOR         = 0x07
	# RFC 1997
	COMMUNITY          = 0x08
	#RFC4456
	ORIGINATOR_ID      = 0x09
	CLUSTER_LIST       = 0x0A
	# RFC 4360
	EXTENDED_COMMUNITY = 0x10
	# RFC 4893
	AS4_PATH           = 0x11
	AS4_AGGREGATOR     = 0x12
	# RFC 4760
	MP_REACH_NLRI      = 0x0e # 14
	MP_UNREACH_NLRI    = 0x0f # 15

	# RFC6514
	PMSI_TUNNEL        = 22

	# RFC5512
	TUNNEL_ENCAP	   = 23

	INTERNAL_WITHDRAWN = 0xFFFD
	INTERNAL_WATCHDOG  = 0xFFFE
	INTERNAL_SPLIT     = 0xFFFF

	def __str__ (self):
		# This should move within the classes and not be here
		if self == 0x01: return "ORIGIN"
		if self == 0x02: return "AS_PATH"
		if self == 0x03: return "NEXT_HOP"
		if self == 0x04: return "MULTI_EXIT_DISC"
		if self == 0x05: return "LOCAL_PREFERENCE"
		if self == 0x06: return "ATOMIC_AGGREGATE"
		if self == 0x07: return "AGGREGATOR"
		if self == 0x08: return "COMMUNITY"
		if self == 0x09: return "ORIGINATOR_ID"
		if self == 0x0A: return "CLUSTER_LIST"
		if self == 0x10: return "EXTENDED_COMMUNITY"
		if self == 0x11: return "AS4_PATH"
		if self == 0x0e: return "MP_REACH_NLRI"
		if self == 0x0f: return "MP_UNREACH_NLRI"
		if self == 22:   return "PMSI_TUNNEL"
		if self == 23:   return "TUNNEL_ENCAP"
		if self == 0xffff: return "INTERNAL SPLIT"
		return 'UNKNOWN ATTRIBUTE (%s)' % hex(self)





# encoding: utf-8
"""
address.py

Created by Thomas Mangin on 2010-01-19.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

from struct import pack

# =================================================================== AFI

# http://www.iana.org/assignments/address-family-numbers/
class AFI (int):
	ipv4 = 0x01
	ipv6 = 0x02
	l2vpn = 0x19

	def __str__ (self):
		if self == 0x01: return "IPv4"
		if self == 0x02: return "IPv6"
		if self == 0x19: return "L2VPN"
		return "unknown afi %d" % self

	def __repr__ (self):
		return str(self)

	def pack (self):
		return pack('!H',self)

# =================================================================== SAFI

# http://www.iana.org/assignments/safi-namespace
class SAFI (int):
	unicast = 1                 # [RFC4760]
	multicast = 2               # [RFC4760]
#	deprecated = 3              # [RFC4760]
	nlri_mpls = 4               # [RFC3107]
#	mcast_vpn = 5               # [draft-ietf-l3vpn-2547bis-mcast-bgp] (TEMPORARY - Expires 2008-06-19)
#	pseudowire = 6              # [draft-ietf-pwe3-dynamic-ms-pw] (TEMPORARY - Expires 2008-08-23) Dynamic Placement of Multi-Segment Pseudowires
#	encapsulation = 7           # [RFC5512]
#
#	tunel = 64                  # [Nalawade]
#	vpls = 65                   # [RFC4761]
#	bgp_mdt = 66                # [Nalawade]
#	bgp_4over6 = 67             # [Cui]
#	bgp_6over4 = 67             # [Cui]
#	vpn_adi = 69                # [RFC-ietf-l1vpn-bgp-auto-discovery-05.txt]

	evpn = 70                   #  draft-ietf-l2vpn-evpn]

	mpls_vpn = 128              # [RFC4364]
#	mcast_bgp_mpls_vpn = 129    # [RFC2547]
	rtc = 132                    # [RFC4684]
	flow_ipv4 = 133             # [RFC5575]
	flow_vpnv4 = 134            # [RFC5575]
#
#	vpn_ad = 140                # [draft-ietf-l3vpn-bgpvpn-auto]
#
#	private = [_ for _ in range(241,254)]   # [RFC4760]
#	unassigned = [_ for _ in range(8,64)] + [_ for _ in range(70,128)]
#	reverved = [0,3] + [130,131] + [_ for _ in range(135,140)] + [_ for _ in range(141,241)] + [255,]    # [RFC4760]

	def __str__ (self):
		if self == 0x01: return "unicast"
		if self == 0x02: return "multicast"
		if self == 0x04: return "nlri-mpls"
		if self == 0x46: return "evpn"
		if self == 0x80: return "mpls-vpn"
		if self == 0x84: return "rtc"
		if self == 0x85: return "flow-ipv4"
		if self == 0x86: return "flow-vpnv4"
		return "unknown safi %d" % self

	def __repr__ (self):
		return str(self)

	def pack (self):
		return chr(self)

## =================================================================== Address

class Address (object):
	def __init__ (self,afi,safi):
		self.afi = AFI(afi)
		self.safi = SAFI(safi)

	def __str__ (self):
		return "%s %s" % (str(self.afi),str(self.safi))

	def __repr__ (self):
		return str(self)



# encoding: utf-8
"""
flag.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

# =================================================================== Flag

class Flag (int):
	EXTENDED_LENGTH = 0x10 # 16
	PARTIAL         = 0x20 # 32
	TRANSITIVE      = 0x40 # 64
	OPTIONAL        = 0x80 # 128

	def __str__ (self):
		r = []
		v = int(self)
		if v & 0x10:
			r.append("EXTENDED_LENGTH")
			v -= 0x10
		if v & 0x20:
			r.append("PARTIAL")
			v -= 0x20
		if v & 0x40:
			r.append("TRANSITIVE")
			v -= 0x40
		if v & 0x80:
			r.append("OPTIONAL")
			v -= 0x80
		if v:
			r.append("UNKNOWN %s" % hex(v))
		return " ".join(r)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import socket

from bagpipe.bgp.common import utils
from bagpipe.bgp.common import logDecorator

from bagpipe.bgp.vpn.vpn_instance import VPNInstance

from bagpipe.bgp.engine import RouteEvent

from bagpipe.bgp.vpn.dataplane_drivers import DummyDataplaneDriver \
    as _DummyDataplaneDriver

from bagpipe.bgp.common.looking_glass import LookingGlass, LGMap

from bagpipe.exabgp.structure.vpn import RouteDistinguisher, VPNLabelledPrefix
from bagpipe.exabgp.structure.mpls import LabelStackEntry
from bagpipe.exabgp.structure.address import AFI, SAFI
from bagpipe.exabgp.structure.ip import Inet, Prefix
from bagpipe.exabgp.message.update.route import Route
from bagpipe.exabgp.message.update.attribute.nexthop import NextHop
from bagpipe.exabgp.message.update.attribute.communities import ECommunities


class DummyDataplaneDriver(_DummyDataplaneDriver):

    pass


class VRF(VPNInstance, LookingGlass):
    # component managing a VRF:
    # - calling a driver to instantiate the dataplane
    # - registering to receive routes for the needed route targets
    # - calling the driver to setup/update/remove routes in the dataplane
    # - cleanup: calling the driver, unregistering for BGP routes

    type = "ipvpn"
    afi = AFI(AFI.ipv4)
    safi = SAFI(SAFI.mpls_vpn)

    @logDecorator.log
    def __init__(self, *args, **kwargs):
        VPNInstance.__init__(self, *args, **kwargs)
        self.readvertised = set()

    def _routeFrom(self, prefix, label, rd):
        return Route(VPNLabelledPrefix(self.afi, self.safi, prefix, rd,
                                       [LabelStackEntry(label, True)]
                                       ))

    def generateVifBGPRoute(self, macAdress, ipPrefix, prefixLen, label):
        # Generate BGP route and advertise it...
        route = self._routeFrom(Prefix(self.afi, ipPrefix, prefixLen), label,
                                RouteDistinguisher(
                                    RouteDistinguisher.TYPE_IP_LOC, None,
                                    self.bgpManager.getLocalAddress(),
                                    self.instanceId)
                                )
        self.log.debug("route attributes: %s", route.attributes)

        return self._newRouteEntry(self.afi, self.safi, self.exportRTs,
                                   route.nlri, route.attributes)

    def _getLocalLabels(self):
        for portData in self.macAddress2LocalPortData.itervalues():
            yield portData['label']

    def _getRDFromLabel(self, label):
        # FIXME: this is a crude hack that will break beyond 10000 VRFs
        return RouteDistinguisher(RouteDistinguisher.TYPE_IP_LOC, None,
                                  self.bgpManager.getLocalAddress(),
                                  10000+label)

    def _routeForReAdvertisement(self, prefix, label):
        route = self._routeFrom(prefix, label,
                                self._getRDFromLabel(label))

        nh = Inet(1, socket.inet_pton(socket.AF_INET,
                  self.dataplane.driver.getLocalAddress()))

        route.attributes.add(NextHop(nh))

        route.attributes.add(ECommunities(self.readvertiseToRTs))

        routeEntry = self._newRouteEntry(self.afi, self.safi,
                                         self.readvertiseToRTs,
                                         route.nlri, route.attributes)
        return routeEntry

    @logDecorator.log
    def _readvertise(self, nlri):
        self.log.debug("Start re-advertising %s from VRF", nlri.prefix)
        for label in self._getLocalLabels():
            self.log.debug("Start re-advertising %s from VRF, with label %s",
                           nlri.prefix, label)
            # need a distinct RD for each route...
            routeEntry = self._routeForReAdvertisement(nlri.prefix, label)
            self._pushEvent(RouteEvent(RouteEvent.ADVERTISE, routeEntry))

        self.readvertised.add(nlri.prefix)

    @logDecorator.log
    def _readvertiseStop(self, nlri):
        self.log.debug("Stop re-advertising %s from VRF", nlri.prefix)
        for label in self._getLocalLabels():
            self.log.debug("Stop re-advertising %s from VRF, with label %s",
                           nlri.prefix, label)
            routeEntry = self._routeForReAdvertisement(nlri.prefix, label)
            self._pushEvent(RouteEvent(RouteEvent.WITHDRAW, routeEntry))

        self.readvertised.remove(nlri.prefix)

    def vifPlugged(self, macAddress, ipAddressPrefix, localPort,
                   advertiseSubnet):
        VPNInstance.vifPlugged(self, macAddress, ipAddressPrefix, localPort,
                               advertiseSubnet)

        label = self.macAddress2LocalPortData[macAddress]['label']
        for prefix in self.readvertised:
            self.log.debug("Re-advertising %s with this port as next hop",
                           prefix)
            routeEntry = self._routeForReAdvertisement(prefix, label)
            self._pushEvent(RouteEvent(RouteEvent.ADVERTISE, routeEntry))

    def vifUnplugged(self, macAddress, ipAddressPrefix, advertiseSubnet):
        label = self.macAddress2LocalPortData[macAddress]['label']
        for prefix in self.readvertised:
            self.log.debug("Stop re-advertising %s with this port as next hop",
                           prefix)
            routeEntry = self._routeForReAdvertisement(prefix, label)
            self._pushEvent(RouteEvent(RouteEvent.WITHDRAW, routeEntry))

        VPNInstance.vifUnplugged(self, macAddress, ipAddressPrefix,
                                 advertiseSubnet)

    # Callbacks for BGP route updates (TrackerWorker) ########################

    def _route2trackedEntry(self, route):
        if isinstance(route.nlri, VPNLabelledPrefix):
            return route.nlri.prefix
        else:
            self.log.error("We should not receive routes of type %s",
                           type(route.nlri))
            return None

    def _toReadvertise(self, route):
        return (len(set(route.routeTargets).intersection(
                    set(self.readvertiseFromRTs))) > 0)

    def _imported(self, route):
        return (len(set(route.routeTargets).intersection(
                    set(self.importRTs))) > 0)

    @utils.synchronized
    @logDecorator.log
    def _newBestRoute(self, entry, newRoute):

        prefix = entry

        if self.readvertise:
            # check if this is a route we need to re-advertise
            self.log.debug("route RTs: %s", newRoute.routeTargets)
            self.log.debug("readv from RTs: %s", self.readvertiseFromRTs)
            if self._toReadvertise(newRoute):
                self.log.debug("Need to re-advertise %s", prefix)
                self._readvertise(newRoute.nlri)
                if not self._imported(newRoute):
                    self.log.debug("No need to setup dataplane for:%s", prefix)
                    return

        encaps = self._checkEncaps(newRoute)
        if not encaps:
            return

        self.dataplane.setupDataplaneForRemoteEndpoint(
            prefix, newRoute.attributes.get(NextHop.ID).next_hop,
            newRoute.nlri.labelStack[0].labelValue, newRoute.nlri, encaps)

    @utils.synchronized
    @logDecorator.log
    def _bestRouteRemoved(self, entry, oldRoute, last):

        prefix = entry

        if self.readvertise and last:
            # check if this is a route we were re-advertising
            if self._toReadvertise(oldRoute):
                self.log.debug("Need to stop re-advertising %s", prefix)
                self._readvertiseStop(oldRoute.nlri)
                if not self._imported(oldRoute):
                    self.log.debug("No need to setup dataplane for:%s", prefix)
                    return

        if self._skipRouteRemoval(last):
            self.log.debug("Skipping removal of non-last route because "
                           "dataplane does not want it")
            return

        self.dataplane.removeDataplaneForRemoteEndpoint(
            prefix, oldRoute.attributes.get(NextHop.ID).next_hop,
            oldRoute.nlri.labelStack[0].labelValue, oldRoute.nlri)

    def getLGMap(self):
        return {
            "readvertised":  (LGMap.VALUE, [repr(prefix) for prefix in
                                            self.readvertised])
        }



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

import socket

from copy import copy

from threading import Thread
from threading import Lock

from netaddr.ip import IPNetwork
import netaddr

from bagpipe.bgp.common import utils
from bagpipe.bgp.common import logDecorator
from bagpipe.bgp.common.looking_glass import LookingGlassLocalLogger, LGMap

from bagpipe.bgp.engine.tracker_worker import TrackerWorker, \
    compareECMP, compareNoECMP

from bagpipe.bgp.engine import RouteEvent, RouteEntry

from bagpipe.exabgp.structure.address import AFI, SAFI

from bagpipe.exabgp.message.update.attribute.communities import ECommunities, \
    Encapsulation
from bagpipe.exabgp.message.update.attribute.id import AttributeID

from bagpipe.exabgp.message.update.attribute.nexthop import NextHop
from bagpipe.exabgp.structure.ip import Inet

from bagpipe.bgp.rest_api import APIException


class VPNInstance(TrackerWorker, Thread, LookingGlassLocalLogger):
    __metaclass__ = ABCMeta

    afi = None
    safi = None

    @logDecorator.log
    def __init__(self, bgpManager, labelAllocator, dataplaneDriver,
                 externalInstanceId, instanceId, importRTs, exportRTs,
                 gatewayIP, mask, readvertise, fallback=None, **kwargs):

        self.instanceType = self.__class__.__name__
        self.instanceId = instanceId

        Thread.__init__(self)
        self.setDaemon(True)

        if dataplaneDriver.ecmpSupport:
            compareRoutes = compareECMP
        else:
            compareRoutes = compareNoECMP

        TrackerWorker.__init__(self, bgpManager, "%s-%d" %
                               (self.instanceType, self.instanceId),
                               compareRoutes)

        LookingGlassLocalLogger.__init__(self,
                                         "%s-%d" % (self.instanceType,
                                                    self.instanceId))
        self.lock = Lock()

        self.importRTs = importRTs
        self.exportRTs = exportRTs
        self.externalInstanceId = externalInstanceId
        self.gatewayIP = gatewayIP
        self.mask = mask

        self.fallback = None

        self.afi = self.__class__.afi
        self.safi = self.__class__.safi
        assert(isinstance(self.afi, AFI))
        assert(isinstance(self.safi, SAFI))

        self.dataplaneDriver = dataplaneDriver
        self.labelAllocator = labelAllocator

        self.instanceLabel = self.labelAllocator.getNewLabel(
            "Incoming traffic for %s %d" % (self.instanceType,
                                            self.instanceId))

        self.localPortData = dict()

        # One local port -> List of endpoints (MAC and IP addresses tuple)
        self.localPort2Endpoints = dict()
        # One MAC address -> One local port
        self.macAddress2LocalPortData = dict()
        # One IP address ->  One MAC address
        self.ipAddress2MacAddress = dict()

        self.dataplane = self.dataplaneDriver.initializeDataplaneInstance(
            self.instanceId, self.externalInstanceId,
            self.gatewayIP, self.mask, self.instanceLabel, **kwargs)

        for rt in self.importRTs:
            self._subscribe(self.afi, self.safi, rt)

        if readvertise:
            self.readvertise = True
            try:
                self.readvertiseToRTs = readvertise['to_rt']
            except KeyError:
                raise APIException("'readvertise' specified with no 'to_rt'")
            self.readvertiseFromRTs = readvertise.get('from_rt', [])
            self.log.debug("readvertise enabled, from RT:%s, to %s",
                           self.readvertiseFromRTs, self.readvertiseToRTs)
            for rt in self.readvertiseFromRTs:
                self._subscribe(self.afi, self.safi, rt)
        else:
            self.log.debug("readvertise not enabled")
            self.readvertise = False

        self.dataplane.update_fallback(fallback)

    @utils.synchronized
    def stop(self):
        self._stop()

    @logDecorator.log
    def _stop(self):
        # cleanup BGP subscriptions
        for rt in self.importRTs:
            self._unsubscribe(self.afi, self.safi, rt)

        self.dataplane.cleanup()

        self.labelAllocator.release(self.instanceLabel)

        # this makes sure that the thread will be stopped, and any remaining
        # routes/subscriptions are released:
        TrackerWorker.stop(self)

    @utils.synchronized
    @logDecorator.log
    def stopIfEmpty(self):
        self.log.debug("localPort2Endpoints: %s", self.localPort2Endpoints)
        if self.isEmpty():
            self._stop()
            return True

        return False

    def isEmpty(self):
        return (not self.localPort2Endpoints)

    def hasEnpoint(self, linuxif):
        return (self.localPort2Endpoints.get(linuxif) is not None)

    @logDecorator.log
    def updateRouteTargets(self, newImportRTs, newExportRTs):
        added_import_rt = set(newImportRTs) - set(self.importRTs)
        removed_import_rt = set(self.importRTs) - set(newImportRTs)

        self.log.debug("%s %d - Added Import RTs: %s",
                       self.instanceType, self.instanceId, added_import_rt)
        self.log.debug("%s %d - Removed Import RTs: %s",
                       self.instanceType, self.instanceId, removed_import_rt)

        # Register to BGP with these route targets
        for rt in added_import_rt:
            self._subscribe(self.afi, self.safi, rt)

        # Unregister from BGP with these route targets
        for rt in removed_import_rt:
            self._unsubscribe(self.afi, self.safi, rt)

        # Update import and export route targets
        self.importRTs = newImportRTs

        # Re-advertise all routes with new export RTs
        self.log.debug("Exports RTs: %s -> %s", self.exportRTs, newExportRTs)
        if frozenset(newExportRTs) != frozenset(self.exportRTs):
            self.log.debug("Will re-export routes with new RTs")
            self.exportRTs = newExportRTs
            for routeEntry in self.getWorkerRouteEntries():
                self.log.info("Re-advertising route %s with updated RTs (%s)",
                              routeEntry.nlri, newExportRTs)

                updatedAttributes = copy(routeEntry.attributes)
                del updatedAttributes[AttributeID.EXTENDED_COMMUNITY]
                updatedAttributes.add(self._genExtendedCommunities())

                updatedRouteEntry = self._newRouteEntry(
                    routeEntry.afi, routeEntry.safi, self.exportRTs,
                    routeEntry.nlri, updatedAttributes)
                self.log.debug("   updated route: %s", updatedRouteEntry)

                self._pushEvent(
                    RouteEvent(RouteEvent.ADVERTISE, updatedRouteEntry))

    def update_fallback(self, fallback):
        if self.fallback != fallback and fallback is not None:
            self.log.info("update fallback: %s", fallback)
            self.fallback = fallback
            self.dataplane.update_fallback(fallback)

    def _parseIPAddressPrefix(self, ipAddressPrefix):
        ipAddress = ""
        mask = 0
        try:
            net = IPNetwork(ipAddressPrefix)
            (ipAddress, mask) = (str(net.ip), net.prefixlen)
        except netaddr.core.AddrFormatError as e:
            raise APIException("Bogus IP prefix: %s" % ipAddressPrefix)

        return (ipAddress, mask)

    def _genExtendedCommunities(self):
        ecommunities = ECommunities(copy(self.exportRTs))
        for encap in self.dataplaneDriver.supportedEncaps():
            if not isinstance(encap, Encapsulation):
                raise Exception("dataplaneDriver.supportedEncaps() should "
                                "return a list of Encapsulation objects")

            if encap != Encapsulation(Encapsulation.DEFAULT):
                ecommunities.add(encap)
        # FIXME: si DEFAULT + xxx => adv MPLS
        return ecommunities

    @abstractmethod
    def generateVifBGPRoute(self, macAddress, ipPrefix, prefixLen, label):
        pass

    def synthesizeVifBGPRoute(self, macAddress, ipPrefix, prefixLen, label):
        routeEntry = self.generateVifBGPRoute(macAddress, ipPrefix, prefixLen,
                                              label)
        assert(isinstance(routeEntry, RouteEntry))

        nh = Inet(1, socket.inet_pton(socket.AF_INET,
                                      self.dataplane.driver.getLocalAddress()))
        routeEntry.attributes.add(NextHop(nh))
        routeEntry.attributes.add(self._genExtendedCommunities())

        return routeEntry

    @utils.synchronized
    @logDecorator.logInfo
    def vifPlugged(self, macAddress, ipAddressPrefix, localPort,
                   advertiseSubnet=False):
        # Check if this port has already been plugged
        # - Verify port informations consistency
        if macAddress in self.macAddress2LocalPortData:
            self.log.debug("MAC address already plugged, checking port "
                           "consistency")
            portData = self.macAddress2LocalPortData[macAddress]

            if (portData.get("port_info") != localPort):
                raise APIException("Port information is not consistent. MAC "
                                   "address cannot be bound to two different"
                                   "ports. Previous plug for port %s "
                                   "(%s != %s)" % (localPort['linuxif'],
                                                   portData.get("port_info"),
                                                   localPort))

        # - Verify (MAC address, IP address) tuple consistency
        if ipAddressPrefix in self.ipAddress2MacAddress:
            if self.ipAddress2MacAddress.get(ipAddressPrefix) != macAddress:
                raise APIException("Inconsistent endpoint info: %s already "
                                   "bound to a MAC address different from %s" %
                                   (ipAddressPrefix, macAddress))
            else:
                return

        # Else, plug port on dataplane
        try:
            # Parse address/mask
            (ipPrefix, prefixLen) = self._parseIPAddressPrefix(ipAddressPrefix)

            self.log.debug("Plugging port (%s)", ipPrefix)

            portData = self.macAddress2LocalPortData.get(macAddress, dict())
            if not portData:
                portData['label'] = self.labelAllocator.getNewLabel(
                    "Incoming traffic for %s %d, interface %s, endpoint %s/%s" %
                    (self.instanceType, self.instanceId, localPort['linuxif'],
                     macAddress, ipAddressPrefix)
                )
                portData["port_info"] = localPort

            # Call driver to setup the dataplane for incoming traffic
            self.dataplane.vifPlugged(macAddress, ipPrefix,
                                      localPort, portData['label'])

            if not advertiseSubnet:
                self.log.debug("Will advertise as /32 instead of /%d" %
                               prefixLen)
                prefixLen = 32

            self.log.info("Synthesizing and advertising BGP route for VIF %s "
                          "endpoint (%s, %s/%d)", localPort['linuxif'],
                          macAddress, ipPrefix, prefixLen)
            routeEntry = self.synthesizeVifBGPRoute(macAddress,
                                                    ipPrefix, prefixLen,
                                                    portData['label'])

            self._pushEvent(RouteEvent(RouteEvent.ADVERTISE, routeEntry))

            if localPort['linuxif'] not in self.localPort2Endpoints:
                self.localPort2Endpoints[localPort['linuxif']] = list()

            self.localPort2Endpoints[localPort['linuxif']].append(
                {'mac': macAddress, 'ip': ipAddressPrefix,
                 }
            )
            self.macAddress2LocalPortData[macAddress] = portData
            self.ipAddress2MacAddress[ipAddressPrefix] = macAddress

        except Exception as e:
            self.log.error("Error in vifPlugged: %s", e)
            if localPort['linuxif'] in self.localPort2Endpoints:
                if len(self.localPort2Endpoints[localPort['linuxif']]) > 1:
                    self.localPort2Endpoints[localPort['linuxif']].remove(
                        {'mac': macAddress, 'ip': ipAddressPrefix}
                    )
                else:
                    del self.localPort2Endpoints[localPort['linuxif']]
            if macAddress in self.macAddress2LocalPortData:
                del self.macAddress2LocalPortData[macAddress]
            if ipAddressPrefix in self.ipAddress2MacAddress:
                del self.ipAddress2MacAddress[ipAddressPrefix]

            raise

    @utils.synchronized
    @logDecorator.logInfo
    def vifUnplugged(self, macAddress, ipAddressPrefix,
                     advertiseSubnet=False):
        # Verify port and endpoint (MAC address, IP address) tuple consistency
        portData = self.macAddress2LocalPortData.get(macAddress)
        if (not portData or
                self.ipAddress2MacAddress.get(ipAddressPrefix) != macAddress):
            self.log.error("vifUnplugged called for endpoint (%s, %s), but no "
                           "consistent informations or was not plugged yet",
                           macAddress, ipAddressPrefix)
            raise APIException("No consistent endpoint (%s, %s) informations "
                               "or was not plugged yet, cannot unplug" %
                               (macAddress, ipAddressPrefix))

        # Finding label and local port informations
        label = portData.get('label')
        localPort = portData.get('port_info')
        if (not label or not localPort):
            self.log.error("vifUnplugged called for endpoint (%s, %s), but "
                           "port data (%s, %s) is incomplete",
                           macAddress, ipAddressPrefix, label, localPort)
            raise Exception("Inconsistent informations for port, bug ?")

        if localPort['linuxif'] in self.localPort2Endpoints:
            # Parse address/mask
            (ipPrefix, prefixLen) = self._parseIPAddressPrefix(ipAddressPrefix)

            lastEndpoint = len(self.localPort2Endpoints[localPort['linuxif']
                                                        ]) <= 1

            if not advertiseSubnet:
                self.log.debug("Will advertise as /32 instead of /%d" %
                               prefixLen)
                prefixLen = 32

            self.log.info("Synthesizing and withdrawing BGP route for VIF %s "
                          "endpoint (%s, %s/%d)", localPort['linuxif'],
                          macAddress, ipPrefix, prefixLen)
            routeEntry = self.synthesizeVifBGPRoute(macAddress,
                                                    ipPrefix, prefixLen,
                                                    label)
            self._pushEvent(RouteEvent(RouteEvent.WITHDRAW, routeEntry))

            # Unplug endpoint from data plane
            self.dataplane.vifUnplugged(
                macAddress, ipPrefix, localPort, label, lastEndpoint)

            # Forget data for this port if last endpoint
            if lastEndpoint:
                # Free label to the allocator
                self.labelAllocator.release(label)

                del self.localPort2Endpoints[localPort['linuxif']]
            else:
                self.localPort2Endpoints[localPort['linuxif']].remove(
                    {'mac': macAddress, 'ip': ipAddressPrefix}
                )

            if not lastEndpoint:
                if not any([endpoint['mac'] == macAddress for endpoint
                            in self.localPort2Endpoints[localPort['linuxif']]]
                           ):
                    del self.macAddress2LocalPortData[macAddress]
            else:
                del self.macAddress2LocalPortData[macAddress]

            del self.ipAddress2MacAddress[ipAddressPrefix]
        else:
            self.log.error("vifUnplugged called for endpoint {%s, %s}, but"
                           " port data is incomplete", macAddress,
                           ipAddressPrefix)
            raise Exception("BGP component bug, check its logs")

    def _checkEncaps(self, route):
        '''
        returns a list of encaps supported by both the dataplane driver and the
        advertized route (based on BGP Encapsulation community)

        logs a warning if there is no common encap
        '''
        advEncaps = None
        try:
            advEncaps = filter(lambda ecom: isinstance(ecom, Encapsulation),
                               route.attributes[
                AttributeID.EXTENDED_COMMUNITY].communities
            )
            self.log.debug("Advertized Encaps: %s", advEncaps)
        except KeyError:
            self.log.debug("no encap advertized, let's use default")

        if not advEncaps:
            advEncaps = [Encapsulation(Encapsulation.DEFAULT)]

        goodEncaps = set(advEncaps) & set(
            self.dataplaneDriver.supportedEncaps())

        if not goodEncaps:
            self.log.warning("No encap supported by dataplane driver for route"
                             " %s, advertized: %s, dataplane supports: {%s}",
                             route, advEncaps,
                             ", ".join([repr(encap) for encap in
                                        self.dataplaneDriver.supportedEncaps()]
                                       ))

        return goodEncaps

    def _skipRouteRemoval(self, last):
        '''
        returns true if the removal of the route should be skipped, based
        whether or not the route removed is the last one and depending on
        the desired behavior for the dataplane driver
        '''
        return not last and not (self.dataplaneDriver.makeB4BreakSupport or
                                 self.dataplaneDriver.ecmpSupport)

    # Looking Glass ####

    def getLGMap(self):
        return {
            "external_instance_id":   (LGMap.VALUE, self.externalInstanceId),
            "dataplane":     (LGMap.DELEGATE, self.dataplane),
            "route_targets": (LGMap.SUBITEM, self.getRTs),
            "gateway_ip":    (LGMap.VALUE, self.gatewayIP),
            "subnet_mask":   (LGMap.VALUE, self.mask),
            "instance_dataplane_id": (LGMap.VALUE, self.instanceLabel),
            "ports":         (LGMap.SUBTREE, self.getLGLocalPortData),
            "readvertise":   (LGMap.SUBITEM, self.getLGReadvertise),
            "fallback":      (LGMap.VALUE, self.fallback)
        }

    def getLGLocalPortData(self, pathPrefix):
        r = {}
        for (port, endpoints) in self.localPort2Endpoints.iteritems():
            eps = []
            for endpoint in endpoints:
                eps.append({
                    'label':
                    self.macAddress2LocalPortData[endpoint['mac']]['label'],
                    'macAddress': endpoint['mac'],
                    'ipAddress': endpoint['ip']
                })

            r[port] = {
                'endpoints': eps
            }
        return r

    def getRTs(self):
        return {
            "import": [repr(rt) for rt in self.importRTs],
            "export": [repr(rt) for rt in self.exportRTs]
        }

    def getLGReadvertise(self):
        if self.readvertise:
            return {'from': [repr(rt) for rt in self.readvertiseFromRTs],
                    'to': [repr(rt) for rt in self.readvertiseToRTs]}
        else:
            return {}



# encoding: utf-8
"""
notification.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

from bagpipe.exabgp.message import Failure, Message

# =================================================================== Notification
# A Notification received from our peer.
# RFC 4271 Section 4.5

class Notification (Message,Failure):
	TYPE = chr(0x03)

	_str_code = {
		1 : "Message header error",
		2 : "OPEN message error",
		3 : "UPDATE message error",
		4 : "Hold timer expired",
		5 : "State machine error",
		6 : "Cease"
	}

	_str_subcode = {
		(1,0) : "Unspecific.",
		(1,1) : "Connection Not Synchronized.",
		(1,2) : "Bad Message Length.",
		(1,3) : "Bad Message Type.",

		(2,0) : "Unspecific.",
		(2,1) : "Unsupported Version Number.",
		(2,2) : "Bad Peer AS.",
		(2,3) : "Bad BGP Identifier.",
		(2,4) : "Unsupported Optional Parameter.",
		(2,5) : "Authentication Notification (Deprecated).",
		(2,6) : "Unacceptable Hold Time.",
		# RFC 5492
		(2,7) : "Unsupported Capability",

		# draft-ietf-idr-bgp-multisession-06
		(2,8) : "Grouping Conflict",
		(2,9) : "Grouping Required",
		(2,10) : "Capability Value Mismatch",

		(3,0) : "Unspecific.",
		(3,1) : "Malformed Attribute List.",
		(3,2) : "Unrecognized Well-known Attribute.",
		(3,3) : "Missing Well-known Attribute.",
		(3,4) : "Attribute Flags Error.",
		(3,5) : "Attribute Length Error.",
		(3,6) : "Invalid ORIGIN Attribute.",
		(3,7) : "AS Routing Loop.",
		(3,8) : "Invalid NEXT_HOP Attribute.",
		(3,9) : "Optional Attribute Error.",
		(3,10) : "Invalid Network Field.",
		(3,11) : "Malformed AS_PATH.",

		(4,0) : "Unspecific.",

		(5,0) : "Unspecific.",
		# RFC 6608
		(5,1) : "Receive Unexpected Message in OpenSent State.",
		(5,2) : "Receive Unexpected Message in OpenConfirm State.",
		(5,3) : "Receive Unexpected Message in Established State.",

		(6,0) : "Unspecific.",
		# RFC 4486
		(6,1) : "Maximum Number of Prefixes Reached",
		(6,2) : "Administrative Shutdown",
		(6,3) : "Peer De-configured",
		(6,4) : "Administrative Reset",
		(6,5) : "Connection Rejected",
		(6,6) : "Other Configuration Change",
		(6,7) : "Connection Collision Resolution",
		(6,8) : "Out of Resources",
	}

	def __init__ (self,code,subcode,data=''):
		self.code = code
		self.subcode = subcode
		self.data = data

	def __str__ (self):
		return "%s: %s [%s]" % (self._str_code.get(self.code,'unknown error'), self._str_subcode.get((self.code,self.subcode),'unknown reason'),self.data)

# =================================================================== Notify
# A Notification we need to inform our peer of.

class Notify (Notification):
	def message (self):
		return self._message("%s%s%s" % (chr(self.code),chr(self.subcode),self.data))

# =================================================================== Notify
# We tried to read data when the connection is not established (as it seems select let us do that !)

class NotConnected (Exception):
	pass



# encoding: utf-8
"""
keepalive.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

from bagpipe.exabgp.message import Message

# =================================================================== KeepAlive

class KeepAlive (Message):
	TYPE = chr(0x04)

	def message (self):
		return self._message()

	def __str__ (self):
		return "KEEPALIVE"

